"""Tests for the bump module."""

from pathlib import Path
from textwrap import dedent
from unittest.mock import MagicMock, patch

import pytest

from bumpversion import bump
from bumpversion.exceptions import ConfigurationError
from bumpversion.files import ConfiguredFile
from bumpversion.scm import Git, SCMInfo
from tests.conftest import get_config_data, inside_dir


@pytest.fixture
def mock_context():
    return {"current_version": "1.2.3", "new_version": "1.2.4"}


def test_get_next_version_with_new_version():
    """Passing a new version should return that version."""
    # Arrange
    config, version_config, current_version = get_config_data({"current_version": "0.1.0"})
    version_part = "patch"
    new_version = "1.2.3"
    expected_next_version = version_config.parse("1.2.3")

    # Act
    actual_next_version = bump.get_next_version(current_version, config, version_part, new_version)

    # Assert
    assert actual_next_version == expected_next_version


def test_get_next_version_with_version_part():
    """If a version part is provided, it should return the increment of that part."""
    # Arrange
    config, version_config, current_version = get_config_data({"current_version": "0.1.0"})
    version_part = "major"
    new_version = None
    expected_next_version = version_config.parse("1.0.0")

    # Act
    actual_next_version = bump.get_next_version(current_version, config, version_part, new_version)

    # Assert
    assert actual_next_version == expected_next_version


def test_get_next_version_with_no_arguments():
    # Arrange
    current_version = MagicMock()
    config = MagicMock()
    version_part = None
    new_version = None

    # Act/Assert
    with pytest.raises(ConfigurationError):
        bump.get_next_version(current_version, config, version_part, new_version)


@patch("bumpversion.files.modify_files")
@patch("bumpversion.bump.update_config_file")
def test_do_bump_with_version_part(mock_update_config_file, mock_modify_files):
    # Arrange
    version_part = "major"
    new_version = None
    config, version_config, current_version = get_config_data(
        {"current_version": "1.2.3", "files": [{"filename": "foo.txt"}, {"filename": "bar.txt"}]}
    )
    config.scm_info = SCMInfo()
    dry_run = False

    # Act
    bump.do_bump(version_part, new_version, config, dry_run=dry_run)

    # Assert
    mock_modify_files.assert_called_once()
    mock_update_config_file.assert_called_once()
    assert {f.file_change.filename for f in mock_modify_files.call_args[0][0]} == {
        "foo.txt",
        "bar.txt",
    }
    assert mock_update_config_file.call_args[0][0] is None
    assert mock_update_config_file.call_args[0][1] == config
    assert mock_update_config_file.call_args[0][2] == current_version
    assert mock_update_config_file.call_args[0][3] == current_version.bump(version_part)
    assert mock_update_config_file.call_args[0][5] is False


@patch("bumpversion.files.modify_files")
@patch("bumpversion.bump.update_config_file")
def test_do_bump_with_new_version(mock_update_config_file, mock_modify_files):
    # Arrange
    version_part = None
    new_version = "2.0.0"
    config, version_config, current_version = get_config_data(
        {
            "current_version": "1.2.3",
        }
    )
    config.scm_info = SCMInfo()
    dry_run = True

    # Act
    bump.do_bump(version_part, new_version, config, dry_run=dry_run)

    # Assert
    mock_modify_files.assert_called_once()
    assert mock_modify_files.call_args[0][0] == []
    assert mock_modify_files.call_args[0][1] == current_version
    assert mock_modify_files.call_args[0][2] == version_config.parse(new_version)

    mock_update_config_file.assert_called_once()
    assert mock_update_config_file.call_args[0][0] is None
    assert mock_update_config_file.call_args[0][1] == config
    assert mock_update_config_file.call_args[0][2] == current_version
    assert mock_update_config_file.call_args[0][3] == version_config.parse(new_version)
    assert mock_update_config_file.call_args[0][5] is dry_run


@patch("bumpversion.files.modify_files")
@patch("bumpversion.bump.update_config_file")
def test_do_bump_when_new_equals_current(mock_update_config_file, mock_modify_files, tmp_path: Path):
    """When the new version is the same as the current version, nothing should happen."""

    # Arrange
    version_part = None
    new_version = "1.2.3"

    with inside_dir(tmp_path):
        config, version_config, current_version = get_config_data({"current_version": "1.2.3"})
        # Act
        bump.do_bump(version_part, new_version, config)

    # Assert
    mock_modify_files.assert_not_called()
    mock_update_config_file.assert_not_called()


def test_do_bump_with_no_arguments():
    # Arrange
    version_part = None
    new_version = None
    config, version_config, current_version = get_config_data({"current_version": "1.2.3"})
    config.scm_info = SCMInfo()
    dry_run = False

    # Act/Assert
    with pytest.raises(ConfigurationError):
        bump.do_bump(version_part, new_version, config, dry_run=dry_run)


def test_commit_and_tag_with_no_tool():
    config, version_config, current_version = get_config_data(
        {
            "current_version": "1.2.3",
        }
    )
    config.scm_info = SCMInfo()
    mock_context = MagicMock()

    bump.commit_and_tag(config, None, [], mock_context, False)


def test_commit_and_tag_with_tool(mock_context):
    config, version_config, current_version = get_config_data(
        {"current_version": "1.2.3", "files": [{"filename": "foo.txt"}, {"filename": "bar.txt"}]}
    )
    config.scm_info = SCMInfo()
    config.scm_info.tool = MagicMock(spec=Git)
    configured_files = [ConfiguredFile(file_cfg, version_config) for file_cfg in config.files]
    bump.commit_and_tag(config, None, configured_files, mock_context, False)
    config.scm_info.tool.commit_to_scm.assert_called_once()
    config.scm_info.tool.tag_in_scm.assert_called_once()
    assert set(config.scm_info.tool.commit_to_scm.call_args[0][0]) == {"foo.txt", "bar.txt"}


def test_commit_and_tag_with_config_file(mock_context):
    config, version_config, current_version = get_config_data(
        {"current_version": "1.2.3", "files": [{"filename": "foo.txt"}, {"filename": "bar.txt"}]}
    )
    config.scm_info = SCMInfo()
    config.scm_info.tool = MagicMock(spec=Git)
    configured_files = [ConfiguredFile(file_cfg, version_config) for file_cfg in config.files]
    bump.commit_and_tag(config, Path("pyproject.toml"), configured_files, mock_context, False)
    config.scm_info.tool.commit_to_scm.assert_called_once()
    config.scm_info.tool.tag_in_scm.assert_called_once()
    assert set(config.scm_info.tool.commit_to_scm.call_args[0][0]) == {"foo.txt", "bar.txt", "pyproject.toml"}


def test_key_path_required_for_toml_change(tmp_path: Path, caplog):
    """If the key_path is not provided, the toml file should use standard search and replace."""
    from bumpversion import config
    from bumpversion.version_part import VersionConfig

    # Arrange
    config_path = tmp_path / "pyproject.toml"
    config_path.write_text(
        dedent(
            """
            [project]
            version = "0.1.26"

            [tool.bumpversion]
            current_version = "0.1.26"
            allow_dirty = true
            commit = true

            [[tool.bumpversion.files]]
            filename = "pyproject.toml"
            search = "version = \\"{current_version}\\""
            replace = "version = \\"{new_version}\\""

            [tool.othertool]
            bake_cookies = true
            ignore-words-list = "sugar, salt, flour"
            """
        ),
        encoding="utf-8",
    )

    conf = config.get_configuration(config_file=config_path)

    # Act
    from click.testing import CliRunner, Result
    from bumpversion import cli

    runner: CliRunner = CliRunner()
    with inside_dir(tmp_path):
        result: Result = runner.invoke(
            cli.cli,
            ["bump", "-vv", "minor"],
        )

    if result.exit_code != 0:
        print(caplog.text)
        print("Here is the output:")
        print(result.output)
        import traceback

        print(traceback.print_exception(result.exc_info[1]))

    # Assert
    assert result.exit_code == 0
    assert config_path.read_text() == dedent(
        """
        [project]
        version = "0.2.0"

        [tool.bumpversion]
        current_version = "0.2.0"
        allow_dirty = true
        commit = true

        [[tool.bumpversion.files]]
        filename = "pyproject.toml"
        search = "version = \\"{current_version}\\""
        replace = "version = \\"{new_version}\\""

        [tool.othertool]
        bake_cookies = true
        ignore-words-list = "sugar, salt, flour"
        """
    )



"""Tests of the visualize module."""

from pathlib import Path

from bumpversion.visualize import Border, BOX_CHARS, connection_str, lead_string, labeled_line, visualize
from bumpversion.config import get_configuration


class TestLeadString:
    """Tests of the lead_string function."""

    def test_returns_string_with_bump(self):
        assert lead_string("1.0.0", Border(*BOX_CHARS["light"])) == "1.0.0 ── bump ─"

    def test_returns_blank_string(self):
        assert lead_string("1.0.0", Border(*BOX_CHARS["light"]), blank=True) == "               "


class TestConnectionStr:
    """Tests of the connection_str function."""

    def test_returns_correct_connection_string(self):
        border = Border(*BOX_CHARS["light"])
        assert connection_str(border, has_next=True, has_previous=True) == border.divider_right + border.line
        assert connection_str(border, has_next=True, has_previous=False) == border.divider_down + border.line
        assert connection_str(border, has_next=False, has_previous=True) == border.corner_bottom_left + border.line
        assert connection_str(border, has_next=False, has_previous=False) == border.line * 2


class TestLabeledLine:
    """Tests of the labeled_line function."""

    def test_no_label_fill_when_no_fit_length(self):
        """Without a fit_length, there will be no filler in the line."""
        border = Border(*BOX_CHARS["light"])
        assert labeled_line("major", border, fit_length=None) == " major ─ "

    def test_pads_label_if_fit_length(self):
        """Without a fit_length, there will be no filler in the line."""
        border = Border(*BOX_CHARS["light"])
        assert labeled_line("major", border, fit_length=10) == " major ────── "


class TestVisualize:
    """Tests of the visualize function."""

    def test_outputs_using_default_config(self, tmp_path: Path, capsys):
        """Test that the correct string is returned."""
        config = get_configuration(tmp_path.joinpath("missing.toml"), current_version="1.0.0")
        visualize(config, "1.0.0")
        captured = capsys.readouterr()
        assert captured.out == "\n".join(
            ["1.0.0 ── bump ─┬─ major ─ 2.0.0", "               ├─ minor ─ 1.1.0", "               ╰─ patch ─ 1.0.1\n"]
        )

    def test_indicates_invalid_paths(self, tmp_path: Path, fixtures_path: Path, capsys):
        """Test that the correct string is returned."""
        config_content = fixtures_path.joinpath("basic_cfg.toml").read_text()
        config_path = tmp_path.joinpath(".bumpversion-1.toml")
        config_path.write_text(config_content, encoding="utf-8")
        config = get_configuration(config_path)
        visualize(config, "1.0.0")
        captured = capsys.readouterr()
        assert captured.out == "\n".join(
            [
                "1.0.0 ── bump ─┬─ major ─── 2.0.0-dev",
                "               ├─ minor ─── 1.1.0-dev",
                "               ├─ patch ─── 1.0.1-dev",
                "               ╰─ release ─ invalid: The part has already the maximum value among ['dev', 'gamma'] and cannot be bumped.\n",
            ]
        )



"""Contains methods for finding and reading configuration files."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, MutableMapping, Union

from bumpversion.config.files_legacy import read_ini_file
from bumpversion.ui import get_indented_logger, print_warning

if TYPE_CHECKING:  # pragma: no-coverage
    from bumpversion.config.models import Config
    from bumpversion.versioning.models import Version

logger = get_indented_logger(__name__)

CONFIG_FILE_SEARCH_ORDER = (
    ".bumpversion.cfg",
    ".bumpversion.toml",
    "setup.cfg",
    "pyproject.toml",
)


def find_config_file(explicit_file: Union[str, Path, None] = None) -> Union[Path, None]:
    """
    Find the configuration file, if it exists.

    If no explicit configuration file is passed, it will search in several files to
    find its configuration.

    Args:
        explicit_file: The configuration file to explicitly use.

    Returns:
        The configuration file path
    """
    search_paths = (
        [Path(explicit_file)] if explicit_file else [Path.cwd().joinpath(path) for path in CONFIG_FILE_SEARCH_ORDER]
    )
    return next(
        (
            cfg_file
            for cfg_file in search_paths
            if cfg_file.exists() and "bumpversion]" in cfg_file.read_text(encoding="utf-8")
        ),
        None,
    )


def read_config_file(config_file: Union[str, Path, None] = None) -> Dict[str, Any]:
    """
    Read the configuration file, if it exists.

    If no explicit configuration file is passed, it will search in several files to
    find its configuration.

    Args:
        config_file: The configuration file to explicitly use.

    Returns:
        A dictionary of read key-values
    """
    if not config_file:
        logger.info("No configuration file found.")
        return {}

    config_path = Path(config_file)
    if not config_path.exists():
        logger.info("Configuration file not found: %s.", config_path)
        return {}

    logger.info("Reading config file: %s", config_file)

    if config_path.suffix == ".cfg":
        print_warning("The .cfg file format is deprecated. Please use .toml instead.")
        return read_ini_file(config_path)
    elif config_path.suffix == ".toml":
        return read_toml_file(config_path)
    else:
        logger.info("Unknown config file suffix: %s. Using defaults.", config_path.suffix)
        return {}


def read_toml_file(file_path: Path) -> Dict[str, Any]:
    """
    Parse a TOML file and return the `bumpversion` section.

    Args:
        file_path: The path to the TOML file.

    Returns:
        dict: A dictionary of the `bumpversion` section.
    """
    import tomlkit

    # Load the TOML file
    toml_data = tomlkit.parse(file_path.read_text(encoding="utf-8")).unwrap()

    return toml_data.get("tool", {}).get("bumpversion", {})


def update_config_file(
    config_file: Union[str, Path],
    config: Config,
    current_version: Version,
    new_version: Version,
    context: MutableMapping,
    dry_run: bool = False,
) -> None:
    """
    Update the current_version key in the configuration file.

    Args:
        config_file: The configuration file to explicitly use.
        config: The configuration to use.
        current_version: The current version.
        new_version: The new version.
        context: The context to use for serialization.
        dry_run: True if the update should be a dry run.
    """
    from bumpversion.config.models import FileChange
    from bumpversion.files import DataFileUpdater

    if not config_file:
        logger.info("\n%sNo configuration file found to update.", logger.indent_str)
        return
    else:
        logger.info("\n%sProcessing config file: %s", logger.indent_str, config_file)
    logger.indent()
    config_path = Path(config_file)
    if config_path.suffix != ".toml":
        logger.info("You must have a `.toml` suffix to update the config file: %s.", config_path)
        return

    # TODO: Eventually this should be transformed into another default "files_to_modify" entry
    datafile_config = FileChange(
        filename=str(config_path),
        key_path="tool.bumpversion.current_version",
        search=config.search,
        replace=config.replace,
        regex=config.regex,
        ignore_missing_version=True,
        ignore_missing_file=True,
        serialize=config.serialize,
        parse=config.parse,
    )

    updater = DataFileUpdater(datafile_config, config.version_config.part_configs)
    updater.update_file(current_version, new_version, context, dry_run)
    logger.dedent()



"""Tests of the VersionPart model."""

from freezegun import freeze_time

from bumpversion.versioning.models import VersionComponentSpec
from bumpversion.versioning.functions import ValuesFunction, NumericFunction, CalVerFunction
import pytest

params = {
    "numeric": [{"optional_value": "0", "first_value": "0"}],
    "values": [{"values": ["alpha", "beta", "gamma"]}],
    "calver": [{"calver_format": "{YYYY}.{MM}.{DD}"}],
}


@pytest.fixture
def version_component_config(request):
    """Return a three-part and a two-part version part configuration."""
    return VersionComponentSpec(**request.param)


class TestVersionComponent:
    class TestCreation:
        @pytest.mark.parametrize(
            ["version_component_config"],
            [
                params["numeric"],
                params["values"],
            ],
            ids=["numeric", "values"],
            indirect=True,
        )
        def test_none_value_uses_optional_value(self, version_component_config: VersionComponentSpec):
            """When passing in no value, the optional value is used."""
            vp = version_component_config.create_component()
            assert vp.value == vp.func.optional_value
            assert vp._value is None

        @pytest.mark.parametrize(
            ["version_component_config"],
            [params["calver"]],
            ids=["calver"],
            indirect=True,
        )
        @freeze_time("2020-05-01")
        def test_none_value_uses_current_date(self, version_component_config: VersionComponentSpec):
            """When passing in no value, the current date is used for CalVer components."""
            vp = version_component_config.create_component()
            assert vp.value == "2020.5.1"
            assert vp._value == "2020.5.1"

        def test_config_with_values_selects_values_function(self):
            values = ["0", "1", "2"]
            vp = VersionComponentSpec(values=values).create_component()
            assert isinstance(vp.func, ValuesFunction)

        def test_config_with_calver_selects_calver_function(self):
            vp = VersionComponentSpec(calver_format="{YYYY}.{MM}").create_component()
            assert isinstance(vp.func, CalVerFunction)

        def test_config_defaults_to_numeric_function(self):
            vp = VersionComponentSpec().create_component()
            assert isinstance(vp.func, NumericFunction)

    @pytest.mark.parametrize(
        ["version_component_config"],
        [
            params["numeric"],
            params["values"],
            params["calver"],
        ],
        ids=["numeric", "values", "calver"],
        indirect=True,
    )
    def test_copy_returns_new_version_part(self, version_component_config):
        vp = version_component_config.create_component(version_component_config.first_value)
        vc = vp.copy()
        assert vp.value == vc.value
        assert id(vp) != id(vc)

    @pytest.mark.parametrize(
        ["version_component_config"],
        [params["numeric"]],
        ids=["numeric"],
        indirect=True,
    )
    def test_bump_increments_numeric_value(self, version_component_config):
        vp = version_component_config.create_component(version_component_config.first_value)
        vc = vp.bump()
        assert vc.value == str(int(version_component_config.first_value) + 1)

    @pytest.mark.parametrize(
        ["version_component_config"],
        [params["values"]],
        ids=["values"],
        indirect=True,
    )
    def test_bump_selects_next_value(self, version_component_config):
        vp = version_component_config.create_component(version_component_config.first_value)
        vc = vp.bump()
        assert vc.value == str(version_component_config.values[1])

    @pytest.mark.parametrize(
        ["version_component_config"],
        [
            params["numeric"],
            params["values"],
            params["calver"],
        ],
        ids=["numeric", "values", "calver"],
        indirect=True,
    )
    def test_non_first_value_is_not_optional(self, version_component_config):
        assert not version_component_config.create_component(version_component_config.first_value).bump().is_optional

    @pytest.mark.parametrize(
        ["version_component_config"],
        [params["calver"]],
        ids=["calver"],
        indirect=True,
    )
    def test_calver_is_not_optional(self, version_component_config):
        vc = version_component_config.create_component(version_component_config.first_value)
        assert not vc.is_optional

    @pytest.mark.parametrize(
        ["version_component_config"],
        [params["numeric"], params["values"]],
        ids=["numeric", "values"],
        indirect=True,
    )
    def test_first_value_is_optional(self, version_component_config):
        vc = version_component_config.create_component(version_component_config.first_value)
        assert vc.is_optional

    @pytest.mark.parametrize(
        ["version_component_config"],
        [
            params["numeric"],
            params["values"],
            params["calver"],
        ],
        ids=["numeric", "values", "calver"],
        indirect=True,
    )
    def test_versionparts_with_same_settings_are_equal(self, version_component_config):
        version1 = version_component_config.create_component(version_component_config.first_value)
        version2 = version_component_config.create_component(version_component_config.first_value)
        assert version1 == version2

    @pytest.mark.parametrize(
        ["version_component_config"],
        [
            params["numeric"],
            params["values"],
            params["calver"],
        ],
        ids=["numeric", "values", "calver"],
        indirect=True,
    )
    def test_null_resets_value_to_first_value(self, version_component_config):
        version1 = version_component_config.create_component(version_component_config.first_value)
        version2 = version1.bump().null()
        assert version2 == version1



"""Utilities for handling command aliases."""

from typing import List, Optional

import rich_click as click
from click import Context
from rich_click.rich_group import RichGroup

from bumpversion.ui import print_warning


class AliasedGroup(RichGroup):
    """
    This following example implements a subclass of Group that accepts a prefix for a command.

    If there were a command called ``push``, it would accept ``pus`` as an alias (so long as it was unique)
    """

    def get_command(self, ctx: Context, cmd_name: str) -> Optional[click.Command]:
        """Given a context and a command name, this returns a Command object if it exists or returns None."""
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        matches = [x for x in self.list_commands(ctx) if x.startswith(cmd_name)]
        if not matches:
            args = [cmd_name, *ctx.args]
            new_ctx = self.make_context(ctx.info_name, args, parent=ctx)
            return click.Group.get_command(self, new_ctx, "bump")
        elif len(matches) == 1:
            return click.Group.get_command(self, ctx, matches[0])
        ctx.fail(f"Too many matches: {', '.join(sorted(matches))}")

    def resolve_command(self, ctx: Context, args: List[str]) -> tuple:
        """Find the command and make sure the full command name is returned."""
        # always return the full command name
        original_args = args[:]
        _, cmd, args = super().resolve_command(ctx, args)

        if cmd.name == "bump" and args != original_args:
            if "bump" in original_args:
                original_args.remove("bump")
            else:
                print_warning(
                    "Calling bumpversion without a subcommand is deprecated. "
                    "Please use `bump-my-version bump` instead"
                )
            return cmd.name, cmd, original_args
        return cmd.name, cmd, args



"""Testing fixtures for Pytest."""

import subprocess
from contextlib import contextmanager
from click.testing import CliRunner
from pathlib import Path
from typing import Generator

import pytest


@pytest.fixture
def tests_path() -> Path:
    """Return the path containing the tests."""
    return Path(__file__).parent


@pytest.fixture
def fixtures_path(tests_path: Path) -> Path:
    """Return the path containing the testing fixtures."""
    return tests_path / "fixtures"


@contextmanager
def inside_dir(dirpath: Path) -> Generator:
    """
    Temporarily switch to a specific directory.

    Args:
        dirpath: Path of the directory to switch to
    """
    import os

    old_path = os.getcwd()
    try:
        os.chdir(dirpath)
        yield
    finally:
        os.chdir(old_path)


def get_config_data(overrides: dict) -> tuple:
    """Get the configuration, version_config and version."""
    from bumpversion import config
    from bumpversion.version_part import VersionConfig

    conf = config.get_configuration(config_file="missing", **overrides)
    version_config = VersionConfig(conf.parse, conf.serialize, conf.search, conf.replace, conf.parts)
    version = version_config.parse(conf.current_version)

    return conf, version_config, version


@pytest.fixture
def git_repo(tmp_path: Path) -> Path:
    """Generate a simple temporary git repo and return the path."""
    subprocess.run(["git", "init"], cwd=tmp_path, check=True, capture_output=True)
    return tmp_path


@pytest.fixture
def hg_repo(tmp_path: Path) -> Path:
    """Generate a simple temporary mercurial repo and return the path."""
    subprocess.run(["hg", "init"], cwd=tmp_path, check=True, capture_output=True)
    return tmp_path


@pytest.fixture
def runner() -> CliRunner:
    """Return a CliRunner instance."""

    return CliRunner()



from click.testing import Result

from bumpversion import cli
from bumpversion import __version__


def test_version_displays_library_version(runner):
    """
    Arrange/Act: Run the `version` subcommand.

    Assert: The output matches the library version.
    """
    result: Result = runner.invoke(cli.cli, ["--version"])
    assert __version__ in result.output.strip(), "Version number should match library version."



"""Tests of the VCS module."""

import subprocess
from pathlib import Path

import pytest
from pytest import param, LogCaptureFixture

from bumpversion import scm
from bumpversion.exceptions import DirtyWorkingDirectoryError
from bumpversion.ui import setup_logging
from tests.conftest import get_config_data, inside_dir


def test_git_is_usable(git_repo: Path) -> None:
    """Should return true if git is available, and it is a git repo."""
    with inside_dir(git_repo):
        assert scm.Git.is_usable()


def test_git_is_not_usable(tmp_path: Path) -> None:
    """Should return false if it is not a git repo."""
    with inside_dir(tmp_path):
        assert not scm.Git.is_usable()


def test_git_asserts_not_dirty(git_repo: Path) -> None:
    """If the git repo is clean, assert_notdirty should return true."""
    with inside_dir(git_repo):
        scm.Git.assert_nondirty()


def test_git_asserts_dirty(git_repo: Path) -> None:
    """If the git repo has modified files, assert_notdirty should return false."""
    readme = git_repo.joinpath("readme.md")
    readme.touch()
    with pytest.raises(DirtyWorkingDirectoryError):
        with inside_dir(git_repo):
            subprocess.run(["git", "add", "readme.md"])
            scm.Git.assert_nondirty()


def test_git_latest_tag_info(git_repo: Path) -> None:
    """Should return information about the latest tag."""
    readme = git_repo.joinpath("readme.md")
    readme.touch()
    tag_prefix = "app/"
    parse_pattern = r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"
    tag_name = f"{tag_prefix}{{new_version}}"
    with inside_dir(git_repo):
        assert scm.Git.latest_tag_info(tag_name, parse_pattern=parse_pattern) == scm.SCMInfo(tool=scm.Git)

        # Add a file and tag
        subprocess.run(["git", "add", "readme.md"])
        subprocess.run(["git", "commit", "-m", "first"])
        subprocess.run(["git", "tag", f"{tag_prefix}0.1.0"])
        tag_info = scm.Git.latest_tag_info(tag_name, parse_pattern=parse_pattern)
        assert tag_info.commit_sha is not None
        assert tag_info.current_version == "0.1.0"
        assert tag_info.distance_to_latest_tag == 0

        # Make it dirty
        git_repo.joinpath("something.md").touch()
        subprocess.run(["git", "add", "something.md"])
        tag_info = scm.Git.latest_tag_info(tag_name, parse_pattern=parse_pattern)
        assert tag_info.commit_sha is not None
        assert tag_info.current_version == "0.1.0"
        assert tag_info.distance_to_latest_tag == 0
        assert tag_info.dirty is True


def test_git_detects_existing_tag(git_repo: Path, caplog: LogCaptureFixture) -> None:
    """Attempting to tag when a tag exists will do nothing."""
    # Arrange
    git_repo.joinpath("readme.md").touch()
    git_repo.joinpath("something.md").touch()
    overrides = {"current_version": "0.1.0", "commit": True, "tag": True}
    context = {
        "current_version": "0.1.0",
        "new_version": "0.2.0",
    }
    setup_logging(2)
    with inside_dir(git_repo):
        conf, version_config, current_version = get_config_data(overrides)
        subprocess.run(["git", "add", "readme.md"])
        subprocess.run(["git", "commit", "-m", "first"])
        subprocess.run(["git", "tag", "v0.1.0"])
        subprocess.run(["git", "add", "something.md"])
        subprocess.run(["git", "commit", "-m", "second"])
        subprocess.run(["git", "tag", "v0.2.0"])

        # Act
        scm.Git.tag_in_scm(config=conf, context=context)

    # Assert
    assert "Will not tag" in caplog.text


def test_hg_is_not_usable(tmp_path: Path) -> None:
    """Should return false if it is not a mercurial repo."""
    with inside_dir(tmp_path):
        assert not scm.Mercurial.is_usable()


def test_hg_is_usable(hg_repo: Path) -> None:
    """Should return false if it is not a mercurial repo."""
    with inside_dir(hg_repo):
        assert scm.Mercurial.is_usable()


@pytest.mark.parametrize(
    ["repo", "scm_command", "scm_class"],
    [
        param("git_repo", "git", scm.Git, id="git"),
        param("hg_repo", "hg", scm.Mercurial, id="hg"),
    ],
)
def test_commit_and_tag_from_below_scm_root(repo: str, scm_command: str, scm_class: scm.SourceCodeManager, request):
    # Arrange
    repo_path: Path = request.getfixturevalue(repo)
    version_path = repo_path / "VERSION"
    version_path.write_text("30.0.3", encoding="utf-8")
    sub_dir_path = repo_path / "subdir"
    sub_dir_path.mkdir(exist_ok=True)

    overrides = {"current_version": "30.0.3", "commit": True, "tag": True, "files": [{"filename": str(version_path)}]}
    context = {
        "current_version": "30.0.3",
        "new_version": "30.1.0",
    }
    with inside_dir(repo_path):
        conf, version_config, current_version = get_config_data(overrides)
        subprocess.run([scm_command, "add", "VERSION"], check=True, capture_output=True)
        subprocess.run([scm_command, "commit", "-m", "initial commit"], check=True, capture_output=True)
        with inside_dir(sub_dir_path):
            version_path.write_text("30.1.0", encoding="utf-8")

            # Act
            scm_class.commit_to_scm(files=[version_path], config=conf, context=context)
            scm_class.tag_in_scm(config=conf, context=context)

        # Assert
        tag_info = scm_class.latest_tag_info(conf.tag_name, parse_pattern=conf.parse)
        if scm_command == "git":
            assert tag_info.commit_sha is not None
            assert tag_info.distance_to_latest_tag == 0
        assert tag_info.current_version == "30.1.0"
        assert tag_info.dirty is False


# write tests for no-commit, no-tag and dry-run
@pytest.mark.parametrize(
    ["repo", "scm_command", "scm_class"],
    [
        param("git_repo", "git", scm.Git, id="git"),
        param("hg_repo", "hg", scm.Mercurial, id="hg"),
    ],
)
@pytest.mark.parametrize(
    ["commit", "tag", "dry_run", "should_commit", "should_tag"],
    [
        param(True, True, True, False, False, id="dry-run-stops-commit-and-tag"),
        param(True, False, False, True, False, id="commit-no-tag"),
        param(False, True, False, False, True, id="no-commit-will-tag"),
    ],
)
def test_commit_tag_dry_run_interactions(
    repo: str,
    scm_command: str,
    scm_class: scm.SourceCodeManager,
    commit: bool,
    tag: bool,
    dry_run: bool,
    should_commit: bool,
    should_tag: bool,
    caplog: LogCaptureFixture,
    request,
):
    """Combinations of commit, tag, dry-run and should produce the expected results."""
    # Arrange
    repo_path: Path = request.getfixturevalue(repo)
    version_path = repo_path / "VERSION"
    version_path.write_text("30.0.3", encoding="utf-8")

    overrides = {"current_version": "30.0.3", "commit": commit, "tag": tag, "files": [{"filename": str(version_path)}]}
    context = {
        "current_version": "30.0.3",
        "new_version": "30.1.0",
    }
    with inside_dir(repo_path):
        conf, version_config, current_version = get_config_data(overrides)
        subprocess.run([scm_command, "add", "VERSION"], check=True, capture_output=True)
        subprocess.run([scm_command, "commit", "-m", "initial commit"], check=True, capture_output=True)
        version_path.write_text("30.1.0", encoding="utf-8")
        # Act
        scm_class.commit_to_scm(files=[version_path], config=conf, context=context, dry_run=dry_run)
        scm_class.tag_in_scm(config=conf, context=context, dry_run=dry_run)

        # Assert
        if commit and dry_run:
            assert "Would commit" in caplog.text
        elif should_commit:
            assert "Committing " in caplog.text
        else:
            assert "Would not commit" in caplog.text

        if tag and dry_run:
            assert "Would tag" in caplog.text
        elif should_tag:
            assert "Tagging " in caplog.text
        else:
            assert "Would not tag" in caplog.text



"""Tests for the autocast module."""

import pytest
from pytest import param

from bumpversion import autocast


@pytest.mark.parametrize(
    "value, expected",
    [
        param("true", True, id="true"),
        param("false", False, id="false"),
        param("True", True, id="True"),
        param("False", False, id="False"),
    ],
)
def test_boolify_valid(value: str, expected: bool) -> None:
    """Test the boolify function."""
    assert autocast.boolify(value) == expected


@pytest.mark.parametrize(
    "value",
    [
        param(1, id="int"),
        param("string", id="str"),
    ],
)
def test_boolify_invalid(value):
    """Test the boolify function."""
    with pytest.raises(ValueError):
        autocast.boolify(value)


def test_noneify():
    """noneify should return None if the string is None."""
    assert autocast.noneify("None") is None
    with pytest.raises(ValueError):
        autocast.noneify("0")


@pytest.mark.parametrize(
    "value,expected",
    [
        param("1,2,3,4", [1, 2, 3, 4], id="int"),
        param("1\n2\n3\n4", [1, 2, 3, 4], id="int"),
        param("s,t,r", ["s", "t", "r"], id="str"),
        param("s,\nt,\nr\n", ["s,", "t,", "r"], id="str_newline"),
        param("1.1,2,3.14", [1.1, 2.0, 3.14], id="float"),
        param("True,False,true,false", [True, False, True, False], id="bool"),
        param("None,None,None", [None, None, None], id="none"),
        param("\nNone\nNone\nNone\n", [None, None, None], id="none with extra newlines"),
    ],
)
def test_listify_valid(value, expected):
    """Test the listify function."""
    assert autocast.listify(value) == expected


def test_listify_invalid():
    """Every element of the list should be the same type."""
    with pytest.raises(TypeError):
        autocast.listify("True,1,None")

    with pytest.raises(ValueError):
        autocast.listify("I am not a list")


@pytest.mark.parametrize(
    "value,expected",
    [
        param("true", True, id="true"),
        param("false", False, id="false"),
        param("True", True, id="True"),
        param("False", False, id="False"),
        param("1,2,3,4", [1, 2, 3, 4], id="int-list"),
        param("s,t,r", ["s", "t", "r"], id="str-list"),
        param("s,\nt,\nr\n", ["s,", "t,", "r"], id="str-list-newline"),
        param("1", 1, id="int"),
        param("1.0", 1.0, id="float"),
        param(1, 1, id="real-int"),
    ],
)
def test_autocast_value(value, expected):
    """Test basic functionality of autocast_value."""
    assert autocast.autocast_value(value) == expected



from pathlib import Path

import pytest
from click import UsageError

from bumpversion import exceptions
from bumpversion.context import get_context
from tests.conftest import get_config_data, inside_dir


def test_bump_version_missing_part():
    overrides = {
        "current_version": "1.0.0",
    }
    conf, version_config, current_version = get_config_data(overrides)
    with pytest.raises(exceptions.InvalidVersionPartError, match="No part named 'bugfix'"):
        current_version.bump("bugfix")


def test_build_number_configuration():
    overrides = {
        "current_version": "2.1.6-5123",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)\-(?P<build>\d+)",
        "serialize": ["{major}.{minor}.{patch}-{build}"],
        "parts": {
            "build": {
                "independent": True,
            }
        },
    }
    conf, version_config, current_version = get_config_data(overrides)

    build_version_bump = current_version.bump("build")
    assert build_version_bump["build"].value == "5124"
    assert build_version_bump["build"].is_independent

    major_version_bump = build_version_bump.bump("major")
    assert major_version_bump["build"].value == "5124"
    assert build_version_bump["build"].is_independent
    assert major_version_bump["major"].value == "3"

    build_version_bump = major_version_bump.bump("build")
    assert build_version_bump["build"].value == "5125"
    assert build_version_bump["build"].is_independent


def test_serialize_with_distance_to_latest_tag():
    """Using ``distance_to_latest_tag`` in the serialization string outputs correctly."""
    from bumpversion.scm import Git, SCMInfo

    overrides = {
        "current_version": "19.6.0",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*",
        "serialize": ["{major}.{minor}.{patch}-pre{distance_to_latest_tag}"],
    }
    conf, version_config, current_version = get_config_data(overrides)
    conf.scm_info = SCMInfo(
        tool=Git, commit_sha="1234123412341234", distance_to_latest_tag=3, current_version="19.6.0", dirty=False
    )
    assert version_config.serialize(current_version, get_context(conf)) == "19.6.0-pre3"


def test_serialize_with_environment_var():
    import os

    os.environ["BUILD_NUMBER"] = "567"
    overrides = {
        "current_version": "2.3.4",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+).*",
        "serialize": ["{major}.{minor}.{patch}-pre{$BUILD_NUMBER}"],
    }
    conf, version_config, current_version = get_config_data(overrides)
    assert version_config.serialize(current_version, get_context(conf)) == "2.3.4-pre567"
    del os.environ["BUILD_NUMBER"]


def test_version_part_invalid_regex_exit(tmp_path: Path) -> None:
    """A version part with an invalid regex should raise an exception."""
    # Arrange
    overrides = {
        "current_version": "12",
        "parse": "*kittens*",
    }
    with inside_dir(tmp_path):
        with pytest.raises(UsageError):
            get_config_data(overrides)



"""Tests for the utils module."""

from itertools import combinations

import pytest
from bumpversion import utils


class TestGetNestedValue:
    """Test the get_nested_value function."""

    def test_returns_value(self):
        """It should return the value of the nested key."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.c"
        expected_value = 1

        # Act
        actual_value = utils.get_nested_value(d, path)

        # Assert
        assert actual_value == expected_value

    def test_invalid_path_raises_keyerror(self):
        """It should raise a KeyError if the path is invalid."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.d"

        # Act/Assert
        with pytest.raises(KeyError):
            utils.get_nested_value(d, path)

    def test_non_dict_element_raises_valueerror(self):
        """It should raise a ValueError if an element in the path is not a dict."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.c.d"

        # Act/Assert
        with pytest.raises(ValueError):
            utils.get_nested_value(d, path)


class TestSetNestedValue:
    """Test the set_nested_value function."""

    def test_sets_value(self):
        """It should set the value of the nested key."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.c"
        value = 2
        expected_d = {"a": {"b": {"c": 2}}}

        # Act
        utils.set_nested_value(d, value, path)

        # Assert
        assert d == expected_d

    def test_invalid_path_raises_keyerror(self):
        """It should raise a KeyError if the path is invalid."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.d.e"
        value = 2

        # Act/Assert
        with pytest.raises(KeyError):
            utils.set_nested_value(d, value, path)

    def test_non_dict_element_raises_valueerror(self):
        """It should raise a ValueError if an element in the path is not a dict."""
        # Arrange
        d = {"a": {"b": {"c": 1}}}
        path = "a.b.c.d"
        value = 2

        # Act/Assert
        with pytest.raises(ValueError):
            utils.set_nested_value(d, value, path)


def pytest_generate_tests(metafunc):
    if "flag_permutation" in metafunc.fixturenames:
        flags = [
            "a",
            "i",
            "L",
            "m",
            "s",
            "u",
            "x",
        ]
        permutes = flags[:]
        permutes.extend(["".join(p) for p in combinations(flags, 2)])
        permutes.extend(["".join(p) for p in combinations(flags, 3)])
        permutes.extend(["".join(p) for p in combinations(flags, 4)])
        permutes.extend(["".join(p) for p in combinations(flags, 5)])
        permutes.extend(["".join(p) for p in combinations(flags, 6)])
        permutes.extend(["".join(p) for p in combinations(flags, 7)])
        metafunc.parametrize(
            "flag_permutation",
            permutes,
        )


class TestExtractRegexFlags:
    def test_returns_flags_when_available(self, flag_permutation):
        """It should return the flags when available."""
        # Arrange
        regex = f"(?{flag_permutation})abc"
        expected = ("abc", f"(?{flag_permutation})")

        # Act
        actual_flags = utils.extract_regex_flags(regex)

        # Assert
        assert actual_flags == expected

    def test_returns_empty_string_when_no_flags(self):
        """It should return an empty string when no flags are present."""
        # Arrange
        regex = "abc"
        expected = ("abc", "")

        # Act
        actual_flags = utils.extract_regex_flags(regex)

        # Assert
        assert actual_flags == expected



"""Test configuration parsing."""

import difflib
import json
from pathlib import Path

from click.testing import CliRunner, Result
import pytest
from pytest import LogCaptureFixture, param, TempPathFactory

from bumpversion.context import get_context
from bumpversion import config
from bumpversion.config.files import find_config_file, CONFIG_FILE_SEARCH_ORDER
import bumpversion.config.utils

from tests.conftest import inside_dir, get_config_data


class TestFindConfigFile:
    """Tests for finding the config file."""

    class TestWhenExplictConfigFileIsPassed:
        """Tests for when an explicit config file is passed."""

        def test_returns_path_to_existing_file(self, tmp_path: Path) -> None:
            """If an explicit config file is passed, it should be returned."""
            cfg_file = tmp_path / "bump.toml"
            cfg_file.write_text('[tool.bumpversion]\ncurrent_version = "1.0.0"', encoding="utf-8")
            assert find_config_file(cfg_file) == cfg_file

        def test_returns_none_when_missing_file(self, tmp_path: Path) -> None:
            """If an explicit config file is passed, it should be returned."""
            cfg_file = tmp_path / "bump.toml"
            assert find_config_file(cfg_file) is None

    class TestWhenNoExplicitConfigFileIsPassed:
        """Tests for when no explicit config file is passed."""

        @pytest.mark.parametrize(
            ["cfg_file_name"],
            (param(item, id=item) for item in CONFIG_FILE_SEARCH_ORDER),
        )
        def test_returns_path_to_existing_default_file(self, tmp_path: Path, cfg_file_name: str) -> None:
            """If no explicit config file is passed, it returns the path to an existing expected file."""
            cfg_file = tmp_path / cfg_file_name
            cfg_file.write_text('[tool.bumpversion]\ncurrent_version = "1.0.0"', encoding="utf-8")
            with inside_dir(tmp_path):
                assert find_config_file() == cfg_file

        def test_returns_none_when_missing_file(self, tmp_path: Path) -> None:
            """If an explicit config file is passed, it should be returned."""
            with inside_dir(tmp_path):
                assert find_config_file() is None

        def test_returns_path_to_existing_file_in_correct_order(self, tmp_path: Path) -> None:
            """If no explicit config file is passed, it returns the path to an existing expected file."""
            expected_order = list(CONFIG_FILE_SEARCH_ORDER)[:]  # make a copy so we can mutate it
            cfg_file_paths = [tmp_path / cfg_file_name for cfg_file_name in expected_order]
            for cfg_file in cfg_file_paths:  # create all the files
                cfg_file.write_text('[tool.bumpversion]\ncurrent_version = "1.0.0"', encoding="utf-8")

            with inside_dir(tmp_path):
                while expected_order:
                    expected_file = expected_order.pop(0)  # the top item in the list is the next expected file
                    expected_path = tmp_path / expected_file
                    assert find_config_file() == expected_path
                    expected_path.unlink()  # remove the file so it doesn't get found again


class TestReadConfigFile:
    """Tests for reading the config file."""

    class TestWhenExplictConfigFileIsPassed:
        def test_returns_empty_dict_when_missing_file(
            self, tmp_path_factory: TempPathFactory, caplog: LogCaptureFixture
        ) -> None:
            """If an explicit config file is passed and doesn't exist, it returns an empty dict."""
            caplog.set_level("INFO")
            tmp_path = tmp_path_factory.mktemp("explicit-file-passed-")
            cfg_file = tmp_path / "bump.toml"
            with inside_dir(tmp_path):
                assert config.read_config_file(cfg_file) == {}
                assert "Configuration file not found" in caplog.text

        def test_returns_dict_of_cfg_file(self, fixtures_path: Path) -> None:
            """Files with a .cfg suffix is parsed into a dict and returned."""
            cfg_file = fixtures_path / "basic_cfg.cfg"
            expected = json.loads(fixtures_path.joinpath("basic_cfg_expected.json").read_text())
            assert config.read_config_file(cfg_file) == expected

        def test_returns_dict_of_toml_file(self, fixtures_path: Path) -> None:
            """Files with a .toml suffix is parsed into a dict and returned."""
            cfg_file = fixtures_path / "basic_cfg.toml"
            expected = json.loads(fixtures_path.joinpath("basic_cfg_expected.json").read_text())
            assert config.read_config_file(cfg_file) == expected

        def test_returns_empty_dict_with_unknown_suffix(
            self, tmp_path_factory: TempPathFactory, caplog: LogCaptureFixture
        ) -> None:
            """Files with an unknown suffix return an empty dict."""
            caplog.set_level("INFO")
            tmp_path = tmp_path_factory.mktemp("explicit-file-passed-")
            cfg_file = tmp_path / "basic_cfg.unknown"
            cfg_file.write_text('[tool.bumpversion]\ncurrent_version = "1.0.0"', encoding="utf-8")
            with inside_dir(tmp_path):
                assert config.read_config_file(cfg_file) == {}
            assert "Unknown config file suffix" in caplog.text

    class TestWhenNoConfigFileIsPassed:
        """Tests for when no explicit config file is passed."""

        def test_returns_empty_dict(self, caplog: LogCaptureFixture) -> None:
            """If no explicit config file is passed, it returns an empty dict."""
            caplog.set_level("INFO")
            assert config.read_config_file() == {}
            assert "No configuration file found." in caplog.text


@pytest.mark.parametrize(
    ["conf_file", "expected_file"],
    [
        param("basic_cfg.toml", "basic_cfg_expected.json", id="toml basic cfg"),
    ],
)
def test_read_toml_file(conf_file: str, expected_file: str, fixtures_path: Path) -> None:
    """Parsing the config file should match the expected results."""
    result = bumpversion.config.files.read_toml_file(fixtures_path.joinpath(conf_file))
    expected = json.loads(fixtures_path.joinpath(expected_file).read_text())
    assert result == expected


def test_multiple_config_files(tmp_path: Path):
    """If there are multiple config files, the first one with content wins."""
    setup_cfg = tmp_path / "setup.cfg"
    setup_cfg.write_text("[metadata]\nname: just-a-name\n", encoding="utf-8")
    bumpversion_cfg = tmp_path / ".bumpversion.cfg"
    bumpversion_cfg.write_text("\n", encoding="utf-8")
    pyproject_toml = tmp_path / "pyproject.toml"
    pyproject_toml.write_text(
        "[tool.bumpversion]\n"
        'current_version = "0.10.5"\n'
        'parse = "(?P<major>\\\\d+)\\\\.(?P<minor>\\\\d+)\\\\.(?P<patch>\\\\d+)(\\\\-(?P<release>[a-z]+))?"\n'
        "serialize = [\n"
        '    "{major}.{minor}.{patch}-{release}",\n'
        '    "{major}.{minor}.{patch}"\n'
        "]\n",
        encoding="utf-8",
    )
    with inside_dir(tmp_path):
        cfg_file = bumpversion.config.files.find_config_file()
        cfg = config.get_configuration(cfg_file)

    assert cfg.current_version == "0.10.5"
    assert cfg.parse == "(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)(\\-(?P<release>[a-z]+))?"
    assert cfg.serialize == ("{major}.{minor}.{patch}-{release}", "{major}.{minor}.{patch}")


TOML_EXPECTED_DIFF = (
    "*** \n"
    "--- \n"
    "***************\n"
    "*** 23 ****\n"
    '! current_version = "1.0.0"\n'
    "--- 23 ----\n"
    '! current_version = "1.0.1"\n'
)


@pytest.mark.parametrize(
    [
        "cfg_file_name",
    ],
    [
        ("pyproject.toml",),
    ],
)
def test_update_config_file(tmp_path: Path, cfg_file_name: str, fixtures_path: Path) -> None:
    """
    Make sure only the version string is updated in the config file.
    """
    expected_diff = TOML_EXPECTED_DIFF
    cfg_path = tmp_path / cfg_file_name
    orig_path = fixtures_path / f"basic_cfg{cfg_path.suffix}"
    cfg_path.write_text(orig_path.read_text(), encoding="utf-8")
    original_content = orig_path.read_text().splitlines(keepends=True)
    with inside_dir(tmp_path):
        cfg = config.get_configuration(cfg_path)
        ctx = get_context(cfg)
    current_version = cfg.version_config.parse("1.0.0")
    new_version = cfg.version_config.parse("1.0.1")
    bumpversion.config.files.update_config_file(cfg_path, cfg, current_version, new_version, ctx)
    new_content = cfg_path.read_text().splitlines(keepends=True)
    difference = difflib.context_diff(original_content, new_content, n=0)
    assert "".join(difference) == expected_diff


def test_pep440_config(git_repo: Path, fixtures_path: Path):
    """
    Check the PEP440 config file.
    """
    from bumpversion.context import get_context
    from bumpversion.bump import get_next_version
    from bumpversion import cli
    import subprocess

    # Arrange

    cfg_path = git_repo / "pyproject.toml"
    orig_path = fixtures_path / "pep440.toml"
    cfg_path.write_text(orig_path.read_text(), encoding="utf-8")
    version_path = git_repo / "VERSION"
    version_path.write_text("1.0.0", encoding="utf-8")
    readme_path = git_repo / "README.md"
    runner: CliRunner = CliRunner()

    with inside_dir(git_repo):
        subprocess.run(["git", "add", "VERSION"], check=True, capture_output=True)
        subprocess.run(["git", "commit", "-m", "initial commit"], check=True, capture_output=True)
        subprocess.run(["git", "tag", "v1.0.0"], check=True, capture_output=True)

        cfg = config.get_configuration(cfg_path)
        ctx = get_context(cfg)
        version = cfg.version_config.parse(cfg.current_version)
        next_version = get_next_version(version, cfg, "patch", None)
        next_version_str = cfg.version_config.serialize(next_version, ctx)
        assert next_version_str == "1.0.1"

        subprocess.run(["git", "checkout", "-b", "my-really-LONG-branch_name"], check=True, capture_output=True)
        readme_path.write_text("This is my branch!", encoding="utf-8")
        result: Result = runner.invoke(cli.cli, ["bump", "dev_label", "--no-tag"])
        assert result.exit_code == 0
        cfg = config.get_configuration(cfg_path)
        assert cfg.current_version == "1.0.0.dev0+myreallylongbranchna"


@pytest.mark.parametrize(
    ["glob_pattern", "file_list"],
    [
        param("*.txt", {Path("file1.txt"), Path("file2.txt")}, id="simple-glob"),
        param("**/*.txt", {Path("file1.txt"), Path("file2.txt"), Path("directory/file3.txt")}, id="recursive-glob"),
    ],
)
def test_get_glob_files(glob_pattern: str, file_list: set, fixtures_path: Path):
    """Get glob files should return all the globbed files and nothing else."""
    overrides = {
        "current_version": "1.0.0",
        "parse": r"(?P<major>\d+)\.(?P<minor>\d+)(\.(?P<release>[a-z]+))?",
        "serialize": ["{major}.{minor}.{release}", "{major}.{minor}"],
        "files": [
            {
                "glob": glob_pattern,
            }
        ],
    }
    conf, version_config, current_version = get_config_data(overrides)
    with inside_dir(fixtures_path.joinpath("glob")):
        result = bumpversion.config.utils.resolve_glob_files(conf.files[0])

    assert len(result) == len(file_list)
    for f in result:
        assert Path(f.filename) in file_list


def test_file_overrides_config(fixtures_path: Path):
    """If a file has a different config, it should override the main config."""
    cfg_file = fixtures_path / "file_config_overrides.toml"
    conf = config.get_configuration(cfg_file)
    file_map = {f.filename: f for f in conf.files}
    assert file_map["should_contain_defaults.txt"].parse == conf.parse
    assert file_map["should_contain_defaults.txt"].serialize == conf.serialize
    assert file_map["should_contain_defaults.txt"].search == conf.search
    assert file_map["should_contain_defaults.txt"].replace == conf.replace
    assert file_map["should_contain_defaults.txt"].regex == conf.regex
    assert file_map["should_contain_defaults.txt"].ignore_missing_version == conf.ignore_missing_version

    assert file_map["should_override_search.txt"].parse == conf.parse
    assert file_map["should_override_search.txt"].serialize == conf.serialize
    assert file_map["should_override_search.txt"].search == "**unreleased**"
    assert file_map["should_override_search.txt"].replace == conf.replace
    assert file_map["should_override_search.txt"].regex == conf.regex
    assert file_map["should_override_search.txt"].ignore_missing_version == conf.ignore_missing_version

    assert file_map["should_override_replace.txt"].parse == conf.parse
    assert file_map["should_override_replace.txt"].serialize == conf.serialize
    assert file_map["should_override_replace.txt"].search == conf.search
    assert file_map["should_override_replace.txt"].replace == "**unreleased**"
    assert file_map["should_override_replace.txt"].regex == conf.regex
    assert file_map["should_override_replace.txt"].ignore_missing_version == conf.ignore_missing_version

    assert file_map["should_override_parse.txt"].parse == r"version(?P<major>\d+)"
    assert file_map["should_override_parse.txt"].serialize == conf.serialize
    assert file_map["should_override_parse.txt"].search == conf.search
    assert file_map["should_override_parse.txt"].replace == conf.replace
    assert file_map["should_override_parse.txt"].regex == conf.regex
    assert file_map["should_override_parse.txt"].ignore_missing_version == conf.ignore_missing_version

    assert file_map["should_override_serialize.txt"].parse == conf.parse
    assert file_map["should_override_serialize.txt"].serialize == ("{major}",)
    assert file_map["should_override_serialize.txt"].search == conf.search
    assert file_map["should_override_serialize.txt"].replace == conf.replace
    assert file_map["should_override_serialize.txt"].regex == conf.regex
    assert file_map["should_override_serialize.txt"].ignore_missing_version == conf.ignore_missing_version

    assert file_map["should_override_ignore_missing.txt"].parse == conf.parse
    assert file_map["should_override_ignore_missing.txt"].serialize == conf.serialize
    assert file_map["should_override_ignore_missing.txt"].search == conf.search
    assert file_map["should_override_ignore_missing.txt"].replace == conf.replace
    assert file_map["should_override_ignore_missing.txt"].regex == conf.regex
    assert file_map["should_override_ignore_missing.txt"].ignore_missing_version is False

    assert file_map["should_override_regex.txt"].parse == conf.parse
    assert file_map["should_override_regex.txt"].serialize == conf.serialize
    assert file_map["should_override_regex.txt"].search == conf.search
    assert file_map["should_override_regex.txt"].replace == conf.replace
    assert file_map["should_override_regex.txt"].regex is False
    assert file_map["should_override_regex.txt"].ignore_missing_version == conf.ignore_missing_version



"""A logger adapter that adds an indent to the beginning of each message."""

import logging
from contextvars import ContextVar
from typing import Any, MutableMapping, Optional, Tuple

CURRENT_INDENT = ContextVar("current_indent", default=0)


class IndentedLoggerAdapter(logging.LoggerAdapter):
    """
    Logger adapter that adds an indent to the beginning of each message.

    Parameters:
        logger: The logger to adapt.
        extra: Extra values to add to the logging context.
        depth: The number of `indent_char` to generate for each indent level.
        indent_char: The character or string to use for indenting.
        reset: `True` if the indent level should be reset to zero.
    """

    def __init__(
        self,
        logger: logging.Logger,
        extra: Optional[dict] = None,
        depth: int = 2,
        indent_char: str = " ",
        reset: bool = False,
    ):
        super().__init__(logger, extra or {})
        self._depth = depth
        self._indent_char = indent_char
        if reset:
            self.reset()

    @property
    def current_indent(self) -> int:
        """
        The current indent level.
        """
        return CURRENT_INDENT.get()

    def indent(self, amount: int = 1) -> None:
        """
        Increase the indent level by `amount`.
        """
        CURRENT_INDENT.set(CURRENT_INDENT.get() + amount)

    def dedent(self, amount: int = 1) -> None:
        """
        Decrease the indent level by `amount`.
        """
        CURRENT_INDENT.set(max(0, CURRENT_INDENT.get() - amount))

    def reset(self) -> None:
        """
        Reset the indent level to zero.
        """
        CURRENT_INDENT.set(0)

    @property
    def indent_str(self) -> str:
        """
        The indent string.
        """
        return (self._indent_char * self._depth) * CURRENT_INDENT.get()

    def process(self, msg: str, kwargs: Optional[MutableMapping[str, Any]]) -> Tuple[str, MutableMapping[str, Any]]:
        """
        Process the message and add the indent.

        Args:
            msg: The logging message.
            kwargs: Keyword arguments passed to the logger.

        Returns:
            A tuple containing the message and keyword arguments.
        """
        msg = self.indent_str + msg

        return msg, kwargs



"""Custom exceptions for BumpVersion."""

from typing import Optional

from click import Context, UsageError


class BumpVersionError(UsageError):
    """Custom base class for all BumpVersion exception types."""

    def __init__(self, message: str, ctx: Optional[Context] = None):
        self.message = message
        self.ctx = ctx


class FormattingError(BumpVersionError):
    """We are unable to represent a version required by a format."""

    pass


class MissingValueError(BumpVersionError):
    """A part required for a version format is empty."""

    pass


class DirtyWorkingDirectoryError(BumpVersionError):
    """The working directory is dirty, and it is not allowed."""

    pass


class SignedTagsError(BumpVersionError):
    """The VCS does not support signed tags."""

    pass


class VersionNotFoundError(BumpVersionError):
    """A version number was not found in a source file."""

    pass


class InvalidVersionPartError(BumpVersionError):
    """The specified part (e.g. 'bugfix') was not found."""

    pass


class ConfigurationError(BumpVersionError):
    """A configuration key-value is missing or in the wrong type."""

    pass


class BadInputError(BumpVersionError):
    """User input was bad."""

    pass



"""Test the create function in the config module."""

from pathlib import Path
import pytest

from tests.conftest import inside_dir
from tomlkit import dumps

from bumpversion.config.create import create_configuration, get_defaults_from_dest
from bumpversion.config import DEFAULTS


BASIC_CONFIG = {
    "tool": {
        "bumpversion": {
            "commit": True,
            "tag": True,
            "current_version": "1.0.0",
            "parse": "(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)(\\-(?P<release>[a-z]+))?",
            "serialize": ["{major}.{minor}.{patch}-{release}", "{major}.{minor}.{patch}"],
            "files": [
                {"filename": "setup.py"},
                {"filename": "bumpversion/__init__.py"},
                {"filename": "CHANGELOG.md", "search": "**unreleased**", "replace": "**v{new_version}**"},
            ],
            "parts": {
                "release": {
                    "optional_value": "gamma",
                    "values": ["dev", "gamma"],
                }
            },
        }
    }
}


@pytest.fixture
def default_config() -> dict:
    """The default configuration with the scm_info and parts removed."""
    defaults = DEFAULTS.copy()
    del defaults["scm_info"]
    del defaults["parts"]
    del defaults["files"]
    return defaults


class TestGetDefaultsFromDest:
    """Test the get_defaults_from_dest function."""

    def test_existing_path_returns_existing_config(self, tmp_path: Path, default_config: dict):
        """If the destination exists, the existing config should be returned."""
        dest_path = tmp_path / "pyproject-1.toml"
        dest_path.write_text(dumps(BASIC_CONFIG), encoding="utf-8")

        with inside_dir(tmp_path):
            config, destination_config = get_defaults_from_dest("pyproject-1.toml")

        default_config.update(BASIC_CONFIG["tool"]["bumpversion"])
        del default_config["files"]
        del default_config["parts"]

        assert config == default_config

    def test_missing_path_returns_default_config(self, tmp_path: Path, default_config: dict):
        """If the destination does not exist, the default config should be returned."""
        with inside_dir(tmp_path):
            config, destination_config = get_defaults_from_dest("pyproject-2.toml")

        default_config["current_version"] = "0.1.0"

        assert config == default_config

    def test_existing_path_with_no_config_returns_default_config(self, tmp_path: Path, default_config: dict):
        """If the destination exists but has no config, the default config should be returned."""
        dest_path = tmp_path / "pyproject-3.toml"
        dest_path.write_text(dumps({"tool": {"poetry": {}}}), encoding="utf-8")

        with inside_dir(tmp_path):
            config, destination_config = get_defaults_from_dest("pyproject-3.toml")

        default_config["current_version"] = "0.1.0"

        assert config == default_config

    def test_existing_path_with_project_version_uses_project_version(self, tmp_path: Path, default_config: dict):
        """If the destination exists and has a project version, the project version should be used as current_version."""
        dest_path = tmp_path / "pyproject-4.toml"
        dest_path.write_text(dumps({"tool": {"poetry": {}}, "project": {"version": "1.2.3"}}), encoding="utf-8")

        with inside_dir(tmp_path):
            config, destination_config = get_defaults_from_dest("pyproject-4.toml")

        default_config["current_version"] = "1.2.3"
        assert config == default_config


class TestCreateConfiguration:
    """Test the create_configuration function."""

    def test_destination_without_prompt_returns_default_config(self, tmp_path: Path, default_config: dict):
        """If the destination is provided and no prompt is requested, the default config should be returned."""
        with inside_dir(tmp_path):
            destination_config = create_configuration("pyproject.toml", prompt=False)

        default_config["current_version"] = "0.1.0"
        default_config["commit_args"] = ""

        assert destination_config["tool"]["bumpversion"]["allow_dirty"] == default_config["allow_dirty"]
        assert destination_config["tool"]["bumpversion"]["commit"] == default_config["commit"]
        assert destination_config["tool"]["bumpversion"]["commit_args"] == default_config["commit_args"]
        assert destination_config["tool"]["bumpversion"]["current_version"] == default_config["current_version"]
        assert (
            destination_config["tool"]["bumpversion"]["ignore_missing_files"] == default_config["ignore_missing_files"]
        )
        assert (
            destination_config["tool"]["bumpversion"]["ignore_missing_version"]
            == default_config["ignore_missing_version"]
        )
        assert destination_config["tool"]["bumpversion"]["message"] == default_config["message"]
        assert destination_config["tool"]["bumpversion"]["parse"] == default_config["parse"]
        assert destination_config["tool"]["bumpversion"]["regex"] == default_config["regex"]
        assert destination_config["tool"]["bumpversion"]["replace"] == default_config["replace"]
        assert destination_config["tool"]["bumpversion"]["search"] == default_config["search"]
        assert destination_config["tool"]["bumpversion"]["serialize"] == list(default_config["serialize"])
        assert destination_config["tool"]["bumpversion"]["sign_tags"] == default_config["sign_tags"]
        assert destination_config["tool"]["bumpversion"]["tag"] == default_config["tag"]
        assert destination_config["tool"]["bumpversion"]["tag_message"] == default_config["tag_message"]
        assert destination_config["tool"]["bumpversion"]["tag_name"] == default_config["tag_name"]

    def test_prompt_true_asks_questions_and_updates_result(self, tmp_path: Path, default_config: dict, mocker):
        """If prompt is True, the user should be prompted for the configuration."""
        answer = {
            "current_version": "1.2.3",
            "commit": True,
            "allow_dirty": True,
            "tag": True,
            "commit_args": "--no-verify",
        }
        mocked_questionary = mocker.patch("bumpversion.config.create.questionary", autospec=True)
        mocked_questionary.form.return_value.ask.return_value = answer
        with inside_dir(tmp_path):
            destination_config = create_configuration("pyproject.toml", prompt=True)

        default_config.update(answer)

        assert destination_config["tool"]["bumpversion"]["allow_dirty"] == default_config["allow_dirty"]
        assert destination_config["tool"]["bumpversion"]["commit"] == default_config["commit"]
        assert destination_config["tool"]["bumpversion"]["commit_args"] == default_config["commit_args"]
        assert destination_config["tool"]["bumpversion"]["current_version"] == default_config["current_version"]
        assert (
            destination_config["tool"]["bumpversion"]["ignore_missing_files"] == default_config["ignore_missing_files"]
        )
        assert (
            destination_config["tool"]["bumpversion"]["ignore_missing_version"]
            == default_config["ignore_missing_version"]
        )
        assert destination_config["tool"]["bumpversion"]["message"] == default_config["message"]
        assert destination_config["tool"]["bumpversion"]["parse"] == default_config["parse"]
        assert destination_config["tool"]["bumpversion"]["regex"] == default_config["regex"]
        assert destination_config["tool"]["bumpversion"]["replace"] == default_config["replace"]
        assert destination_config["tool"]["bumpversion"]["search"] == default_config["search"]
        assert destination_config["tool"]["bumpversion"]["serialize"] == list(default_config["serialize"])
        assert destination_config["tool"]["bumpversion"]["sign_tags"] == default_config["sign_tags"]
        assert destination_config["tool"]["bumpversion"]["tag"] == default_config["tag"]
        assert destination_config["tool"]["bumpversion"]["tag_message"] == default_config["tag_message"]
        assert destination_config["tool"]["bumpversion"]["tag_name"] == default_config["tag_name"]
        assert mocked_questionary.form.return_value.ask.call_count == 1



"""Version changing methods."""

import shlex
from pathlib import Path
from typing import TYPE_CHECKING, ChainMap, List, Optional

if TYPE_CHECKING:  # pragma: no-coverage
    from bumpversion.files import ConfiguredFile
    from bumpversion.versioning.models import Version

from bumpversion.config import Config
from bumpversion.config.files import update_config_file
from bumpversion.config.files_legacy import update_ini_config_file
from bumpversion.context import get_context
from bumpversion.exceptions import ConfigurationError
from bumpversion.ui import get_indented_logger
from bumpversion.utils import key_val_string

logger = get_indented_logger(__name__)


def get_next_version(
    current_version: "Version", config: Config, version_part: Optional[str], new_version: Optional[str]
) -> "Version":
    """
    Bump the version_part to the next value.

    Args:
        current_version: The current version
        config: The current configuration
        version_part: Optional part of the version to bump
        new_version: Optional specific version to bump to

    Returns:
        The new version

    Raises:
        ConfigurationError: If it can't generate the next version.
    """
    if new_version:
        logger.info("Attempting to set new version '%s'", new_version)
        logger.indent()
        next_version = config.version_config.parse(new_version)
    elif version_part:
        logger.info("Attempting to increment part '%s'", version_part)
        logger.indent()
        next_version = current_version.bump(version_part)
    else:
        raise ConfigurationError("Unable to get the next version.")

    logger.info("Values are now: %s", key_val_string(next_version.components))
    logger.dedent()
    return next_version


def do_bump(
    version_part: Optional[str],
    new_version: Optional[str],
    config: Config,
    config_file: Optional[Path] = None,
    dry_run: bool = False,
) -> None:
    """
    Bump the version_part to the next value or set the version to new_version.

    Args:
        version_part: The version part to bump
        new_version: The explicit version to set
        config: The configuration to use
        config_file: The configuration file to update
        dry_run: True if the operation should be a dry run
    """
    from bumpversion.files import modify_files, resolve_file_config

    logger.indent()

    ctx = get_context(config)
    logger.info("Parsing current version '%s'", config.current_version)
    logger.indent()
    version = config.version_config.parse(config.current_version)
    logger.dedent()
    next_version = get_next_version(version, config, version_part, new_version)
    next_version_str = config.version_config.serialize(next_version, ctx)
    logger.info("New version will be '%s'", next_version_str)

    if config.current_version == next_version_str:
        logger.info("Version is already '%s'", next_version_str)
        return

    logger.dedent()

    if dry_run:
        logger.info("Dry run active, won't touch any files.")

    ctx = get_context(config, version, next_version)

    configured_files = resolve_file_config(config.files_to_modify, config.version_config)
    modify_files(configured_files, version, next_version, ctx, dry_run)
    if config_file and config_file.suffix in {".cfg", ".ini"}:
        update_ini_config_file(config_file, config.current_version, next_version_str, dry_run)  # pragma: no-coverage
    else:
        update_config_file(config_file, config, version, next_version, ctx, dry_run)

    ctx = get_context(config, version, next_version)
    ctx["new_version"] = next_version_str
    commit_and_tag(config, config_file, configured_files, ctx, dry_run)
    logger.info("Done.")


def commit_and_tag(
    config: Config,
    config_file: Optional[Path],
    configured_files: List["ConfiguredFile"],
    ctx: ChainMap,
    dry_run: bool = False,
) -> None:
    """
    Commit and tag the changes if a tool is configured.

    Args:
        config: The configuration
        config_file: The configuration file to include in the commit, if it exists
        configured_files: A list of files to commit
        ctx: The context used to render the tag and tag message
        dry_run: True if the operation should be a dry run
    """
    if not config.scm_info.tool:
        return

    extra_args = shlex.split(config.commit_args) if config.commit_args else []

    commit_files = {f.file_change.filename for f in configured_files}
    if config_file:
        commit_files |= {str(config_file)}

    config.scm_info.tool.commit_to_scm(list(commit_files), config, ctx, extra_args, dry_run)
    config.scm_info.tool.tag_in_scm(config, ctx, dry_run)



"""Models for managing versioning of software projects."""

from __future__ import annotations

from collections import defaultdict, deque
from itertools import chain
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import BaseModel, model_validator

from bumpversion.exceptions import InvalidVersionPartError
from bumpversion.utils import key_val_string
from bumpversion.versioning.functions import CalVerFunction, NumericFunction, PartFunction, ValuesFunction


class VersionComponent:
    """
    Represent part of a version number.

    Determines the PartFunction that rules how the part behaves when increased or reset
    based on the configuration given.
    """

    def __init__(
        self,
        values: Optional[list] = None,
        optional_value: Optional[str] = None,
        first_value: Union[str, int, None] = None,
        independent: bool = False,
        always_increment: bool = False,
        calver_format: Optional[str] = None,
        source: Optional[str] = None,
        value: Union[str, int, None] = None,
    ):
        self._value = str(value) if value is not None else None
        self.func: Optional[PartFunction] = None
        self.always_increment = always_increment
        self.independent = True if always_increment else independent
        self.source = source
        self.calver_format = calver_format
        if values:
            str_values = [str(v) for v in values]
            str_optional_value = str(optional_value) if optional_value is not None else None
            str_first_value = str(first_value) if first_value is not None else None
            self.func = ValuesFunction(str_values, str_optional_value, str_first_value)
        elif calver_format:
            self.func = CalVerFunction(calver_format)
            self._value = self._value or self.func.first_value
        else:
            self.func = NumericFunction(optional_value, first_value or "0")

    @property
    def value(self) -> str:
        """Return the value of the part."""
        return self._value or self.func.optional_value

    def copy(self) -> "VersionComponent":
        """Return a copy of the part."""
        return VersionComponent(
            values=getattr(self.func, "_values", None),
            optional_value=self.func.optional_value,
            first_value=self.func.first_value,
            independent=self.independent,
            always_increment=self.always_increment,
            calver_format=self.calver_format,
            source=self.source,
            value=self._value,
        )

    def bump(self) -> "VersionComponent":
        """Return a part with bumped value."""
        new_component = self.copy()
        new_component._value = self.func.bump(self.value)
        return new_component

    def null(self) -> "VersionComponent":
        """Return a part with first value."""
        new_component = self.copy()
        new_component._value = self.func.first_value
        return new_component

    @property
    def is_optional(self) -> bool:
        """Is the part optional?"""
        return self.value == self.func.optional_value

    @property
    def is_independent(self) -> bool:
        """Is the part independent of the other parts?"""
        return self.independent

    def __format__(self, format_spec: str) -> str:
        try:
            val = int(self.value)
        except ValueError:
            return self.value
        else:
            return int.__format__(val, format_spec)

    def __repr__(self) -> str:
        return f"<bumpversion.VersionPart:{self.func.__class__.__name__}:{self.value}>"

    def __eq__(self, other: Any) -> bool:
        return self.value == other.value if isinstance(other, VersionComponent) else False


class VersionComponentSpec(BaseModel):
    """
    Configuration of a version component.

    This is used to read in the configuration from the bumpversion config file.
    """

    values: Optional[list] = None
    """The possible values for the component. If it and `calver_format` is None, the component is numeric."""

    optional_value: Optional[str] = None  # Optional.
    """The value that is optional to include in the version.

    - Defaults to first value in values or 0 in the case of numeric.
    - Empty string means nothing is optional.
    - CalVer components ignore this."""

    first_value: Union[str, int, None] = None
    """The first value to increment from."""

    independent: bool = False
    """Is the component independent of the other components?"""

    always_increment: bool = False
    """Should the component always increment, even if it is not necessary?"""

    calver_format: Optional[str] = None
    """The format string for a CalVer component."""

    # source: Optional[str] = None  # Name of environment variable or context variable to use as the source for value
    depends_on: Optional[str] = None
    """The name of the component this component depends on."""

    @model_validator(mode="before")
    @classmethod
    def set_always_increment_with_calver(cls, data: Any) -> Any:
        """Set always_increment to True if calver_format is present."""
        if isinstance(data, dict) and data.get("calver_format"):
            data["always_increment"] = True
        return data

    def create_component(self, value: Union[str, int, None] = None) -> VersionComponent:
        """Generate a version component from the configuration."""
        return VersionComponent(
            values=self.values,
            optional_value=self.optional_value,
            first_value=self.first_value,
            independent=self.independent,
            always_increment=self.always_increment,
            calver_format=self.calver_format,
            # source=self.source,
            value=value,
        )


class VersionSpec:
    """The specification of a version's components and their relationships."""

    def __init__(self, components: Dict[str, VersionComponentSpec], order: Optional[List[str]] = None):
        if not components:
            raise ValueError("A VersionSpec must have at least one component.")
        if not order:
            order = list(components.keys())
        if len(set(order) - set(components.keys())) > 0:
            raise ValueError("The order of components refers to items that are not in your components.")

        self.component_configs = components
        self.order = order
        self.dependency_map = defaultdict(list)
        previous_component = self.order[0]
        self.always_increment = [name for name, config in self.component_configs.items() if config.always_increment]
        for component in self.order[1:]:
            if self.component_configs[component].independent:
                continue
            elif self.component_configs[component].depends_on:
                self.dependency_map[self.component_configs[component].depends_on].append(component)
            else:
                self.dependency_map[previous_component].append(component)
            previous_component = component

    def create_version(self, values: Dict[str, str]) -> "Version":
        """Generate a version from the given values."""
        components = {
            key: comp_config.create_component(value=values.get(key))
            for key, comp_config in self.component_configs.items()
        }
        return Version(version_spec=self, components=components)

    def get_dependents(self, component_name: str) -> List[str]:
        """Return the parts that depend on the given part."""
        stack = deque(self.dependency_map.get(component_name, []), maxlen=len(self.order))
        visited = []

        while stack:
            e = stack.pop()
            if e not in visited:
                visited.append(e)
                stack.extendleft(self.dependency_map[e])

        return visited


class Version:
    """The specification of a version and its parts."""

    def __init__(
        self, version_spec: VersionSpec, components: Dict[str, VersionComponent], original: Optional[str] = None
    ):
        self.version_spec = version_spec
        self.components = components
        self.original = original

    def values(self) -> Dict[str, VersionComponent]:
        """Return the values of the parts."""
        return dict(self.components.items())

    def __getitem__(self, key: str) -> VersionComponent:
        return self.components[key]

    def __len__(self) -> int:
        return len(self.components)

    def __iter__(self):
        return iter(self.components)

    def __repr__(self):
        return f"<bumpversion.Version:{key_val_string(self.components)}>"

    def __eq__(self, other: Any) -> bool:
        return (
            all(value == other.components[key] for key, value in self.components.items())
            if isinstance(other, Version)
            else False
        )

    def required_components(self) -> List[str]:
        """Return the names of the parts that are required."""
        return [key for key, value in self.components.items() if value.value != value.func.optional_value]

    def bump(self, component_name: str) -> "Version":
        """Increase the value of the specified component, reset its dependents, and return a new Version."""
        if component_name not in self.components:
            raise InvalidVersionPartError(f"No part named {component_name!r}")

        new_values = dict(self.components.items())
        always_incr_values, components_to_reset = self._always_increment()
        new_values.update(always_incr_values)

        if component_name not in components_to_reset:
            new_values[component_name] = self.components[component_name].bump()
            components_to_reset |= set(self.version_spec.get_dependents(component_name))

        for component in components_to_reset:
            if not self.components[component].is_independent:
                new_values[component] = self.components[component].null()

        return Version(self.version_spec, new_values, self.original)

    def _always_incr_dependencies(self) -> dict:
        """Return the components that always increment and depend on the given component."""
        return {name: self.version_spec.get_dependents(name) for name in self.version_spec.always_increment}

    def _increment_always_incr(self) -> dict:
        """Increase the values of the components that always increment."""
        components = self.version_spec.always_increment
        return {name: self.components[name].bump() for name in components}

    def _always_increment(self) -> Tuple[dict, set]:
        """Return the components that always increment and their dependents."""
        values = self._increment_always_incr()
        dependents = self._always_incr_dependencies()
        for component_name, value in values.items():
            if value == self.components[component_name]:
                dependents.pop(component_name, None)
        unique_dependents = set(chain.from_iterable(dependents.values()))
        return values, unique_dependents



"""Methods for changing files."""

import os.path
import re
from copy import deepcopy
from difflib import context_diff
from pathlib import Path
from typing import Dict, List, MutableMapping, Optional

from bumpversion.config.models import FileChange
from bumpversion.exceptions import VersionNotFoundError
from bumpversion.ui import get_indented_logger
from bumpversion.utils import get_nested_value, set_nested_value
from bumpversion.version_part import VersionConfig
from bumpversion.versioning.models import Version, VersionComponentSpec

logger = get_indented_logger(__name__)


def contains_pattern(search: re.Pattern, contents: str) -> bool:
    """Does the search pattern match any part of the contents?"""
    if not search or not contents:
        return False

    for m in re.finditer(search, contents):
        line_no = contents.count("\n", 0, m.start(0)) + 1
        logger.info(
            "Found '%s' at line %s: %s",
            search.pattern,
            line_no,
            m.string[m.start() : m.end(0)],
        )
        return True
    return False


def log_changes(file_path: str, file_content_before: str, file_content_after: str, dry_run: bool = False) -> None:
    """
    Log the changes that would be made to the file.

    Args:
        file_path: The path to the file
        file_content_before: The file contents before the change
        file_content_after: The file contents after the change
        dry_run: True if this is a report-only job
    """
    if file_content_before != file_content_after:
        logger.info("%s file %s:", "Would change" if dry_run else "Changing", file_path)
        logger.indent()
        indent_str = logger.indent_str

        logger.info(
            f"\n{indent_str}".join(
                list(
                    context_diff(
                        file_content_before.splitlines(),
                        file_content_after.splitlines(),
                        fromfile=f"before {file_path}",
                        tofile=f"after {file_path}",
                        lineterm="",
                    )
                )
            ),
        )
        logger.dedent()
    else:
        logger.info("%s file %s", "Would not change" if dry_run else "Not changing", file_path)


class ConfiguredFile:
    """A file to modify in a configured way."""

    def __init__(
        self,
        file_change: FileChange,
        version_config: VersionConfig,
        search: Optional[str] = None,
        replace: Optional[str] = None,
    ) -> None:
        replacements = [replace, file_change.replace, version_config.replace]
        replacement = next((r for r in replacements if r is not None), "")
        self.file_change = FileChange(
            parse=file_change.parse or version_config.parse_regex.pattern,
            serialize=file_change.serialize or version_config.serialize_formats,
            search=search or file_change.search or version_config.search,
            replace=replacement,
            regex=file_change.regex or False,
            ignore_missing_version=file_change.ignore_missing_version or False,
            ignore_missing_file=file_change.ignore_missing_file or False,
            filename=file_change.filename,
            glob=file_change.glob,
            key_path=file_change.key_path,
        )
        self.version_config = VersionConfig(
            self.file_change.parse,
            self.file_change.serialize,
            self.file_change.search,
            self.file_change.replace,
            version_config.part_configs,
        )
        self._newlines: Optional[str] = None

    def get_file_contents(self) -> str:
        """
        Return the contents of the file.

        Raises:
            FileNotFoundError: if the file doesn't exist

        Returns:
            The contents of the file
        """
        if not os.path.exists(self.file_change.filename):
            raise FileNotFoundError(f"File not found: '{self.file_change.filename}'")  # pragma: no-coverage

        with open(self.file_change.filename, "rt", encoding="utf-8") as f:
            contents = f.read()
            self._newlines = f.newlines[0] if isinstance(f.newlines, tuple) else f.newlines
            return contents

    def write_file_contents(self, contents: str) -> None:
        """Write the contents of the file."""
        if self._newlines is None:
            _ = self.get_file_contents()

        with open(self.file_change.filename, "wt", encoding="utf-8", newline=self._newlines) as f:
            f.write(contents)

    def _contains_change_pattern(
        self, search_expression: re.Pattern, raw_search_expression: str, version: Version, context: MutableMapping
    ) -> bool:
        """
        Does the file contain the change pattern?

        Args:
            search_expression: The compiled search expression
            raw_search_expression: The raw search expression
            version: The version to check, in case it's not the same as the original
            context: The context to use

        Raises:
            VersionNotFoundError: if the version number isn't present in this file.

        Returns:
            True if the version number is in fact present.
        """
        file_contents = self.get_file_contents()
        if contains_pattern(search_expression, file_contents):
            return True

        # The `search` pattern did not match, but the original supplied
        # version number (representing the same version part values) might
        # match instead. This is probably the case if environment variables are used.

        # check whether `search` isn't customized
        search_pattern_is_default = self.file_change.search == self.version_config.search

        if search_pattern_is_default and contains_pattern(re.compile(re.escape(version.original)), file_contents):
            # The original version is present, and we're not looking for something
            # more specific -> this is accepted as a match
            return True

        # version not found
        if self.file_change.ignore_missing_version:
            return False
        raise VersionNotFoundError(f"Did not find '{raw_search_expression}' in file: '{self.file_change.filename}'")

    def make_file_change(
        self, current_version: Version, new_version: Version, context: MutableMapping, dry_run: bool = False
    ) -> None:
        """Make the change to the file."""
        logger.info(
            "\n%sFile %s: replace `%s` with `%s`",
            logger.indent_str,
            self.file_change.filename,
            self.file_change.search,
            self.file_change.replace,
        )
        logger.indent()
        if not os.path.exists(self.file_change.filename):
            if self.file_change.ignore_missing_file:
                logger.info("File not found, but ignoring")
                logger.dedent()
                return
            raise FileNotFoundError(f"File not found: '{self.file_change.filename}'")  # pragma: no-coverage
        context["current_version"] = self._get_serialized_version("current_version", current_version, context)
        if new_version:
            context["new_version"] = self._get_serialized_version("new_version", new_version, context)
        else:
            logger.debug("No new version, using current version as new version")
            context["new_version"] = context["current_version"]

        search_for, raw_search_pattern = self.file_change.get_search_pattern(context)
        replace_with = self.version_config.replace.format(**context)

        if not self._contains_change_pattern(search_for, raw_search_pattern, current_version, context):
            logger.dedent()
            return

        file_content_before = self.get_file_contents()

        file_content_after = search_for.sub(replace_with, file_content_before)

        if file_content_before == file_content_after and current_version.original:
            og_context = deepcopy(context)
            og_context["current_version"] = current_version.original
            search_for_og, og_raw_search_pattern = self.file_change.get_search_pattern(og_context)
            file_content_after = search_for_og.sub(replace_with, file_content_before)

        log_changes(self.file_change.filename, file_content_before, file_content_after, dry_run)
        logger.dedent()
        if not dry_run:  # pragma: no-coverage
            self.write_file_contents(file_content_after)

    def _get_serialized_version(self, context_key: str, version: Version, context: MutableMapping) -> str:
        """Get the serialized version."""
        logger.debug("Serializing the %s", context_key.replace("_", " "))
        logger.indent()
        serialized_version = self.version_config.serialize(version, context)
        logger.dedent()
        return serialized_version

    def __str__(self) -> str:  # pragma: no-coverage
        return self.file_change.filename

    def __repr__(self) -> str:  # pragma: no-coverage
        return f"<bumpversion.ConfiguredFile:{self.file_change.filename}>"


def resolve_file_config(
    files: List[FileChange], version_config: VersionConfig, search: Optional[str] = None, replace: Optional[str] = None
) -> List[ConfiguredFile]:
    """
    Resolve the files, searching and replacing values according to the FileConfig.

    Args:
        files: A list of file configurations
        version_config: How the version should be changed
        search: The search pattern to use instead of any configured search pattern
        replace: The replace pattern to use instead of any configured replace pattern

    Returns:
        A list of ConfiguredFiles
    """
    return [ConfiguredFile(file_cfg, version_config, search, replace) for file_cfg in files]


def modify_files(
    files: List[ConfiguredFile],
    current_version: Version,
    new_version: Version,
    context: MutableMapping,
    dry_run: bool = False,
) -> None:
    """
    Modify the files, searching and replacing values according to the FileConfig.

    Args:
        files: The list of configured files
        current_version: The current version
        new_version: The next version
        context: The context used for rendering the version
        dry_run: True if this should be a report-only job
    """
    # _check_files_contain_version(files, current_version, context)
    for f in files:
        f.make_file_change(current_version, new_version, context, dry_run)


class FileUpdater:
    """A class to handle updating files."""

    def __init__(
        self,
        file_change: FileChange,
        version_config: VersionConfig,
        search: Optional[str] = None,
        replace: Optional[str] = None,
    ) -> None:
        self.file_change = FileChange(
            parse=file_change.parse or version_config.parse_regex.pattern,
            serialize=file_change.serialize or version_config.serialize_formats,
            search=search or file_change.search or version_config.search,
            replace=replace or file_change.replace or version_config.replace,
            regex=file_change.regex or False,
            ignore_missing_file=file_change.ignore_missing_file or False,
            ignore_missing_version=file_change.ignore_missing_version or False,
            filename=file_change.filename,
            glob=file_change.glob,
            key_path=file_change.key_path,
        )
        self.version_config = VersionConfig(
            self.file_change.parse,
            self.file_change.serialize,
            self.file_change.search,
            self.file_change.replace,
            version_config.part_configs,
        )
        self._newlines: Optional[str] = None

    def update_file(
        self, current_version: Version, new_version: Version, context: MutableMapping, dry_run: bool = False
    ) -> None:
        """Update the files."""
        # TODO: Implement this
        pass


class DataFileUpdater:
    """A class to handle updating files."""

    def __init__(
        self,
        file_change: FileChange,
        version_part_configs: Dict[str, VersionComponentSpec],
    ) -> None:
        self.file_change = file_change
        self.version_config = VersionConfig(
            self.file_change.parse,
            self.file_change.serialize,
            self.file_change.search,
            self.file_change.replace,
            version_part_configs,
        )
        self.path = Path(self.file_change.filename)
        self._newlines: Optional[str] = None

    def update_file(
        self, current_version: Version, new_version: Version, context: MutableMapping, dry_run: bool = False
    ) -> None:
        """Update the files."""
        new_context = deepcopy(context)
        new_context["current_version"] = self.version_config.serialize(current_version, context)
        new_context["new_version"] = self.version_config.serialize(new_version, context)
        search_for, raw_search_pattern = self.file_change.get_search_pattern(new_context)
        replace_with = self.file_change.replace.format(**new_context)
        if self.path.suffix == ".toml":
            self._update_toml_file(search_for, raw_search_pattern, replace_with, dry_run)

    def _update_toml_file(
        self, search_for: re.Pattern, raw_search_pattern: str, replace_with: str, dry_run: bool = False
    ) -> None:
        """Update a TOML file."""
        import tomlkit

        toml_data = tomlkit.parse(self.path.read_text(encoding="utf-8"))
        value_before = get_nested_value(toml_data, self.file_change.key_path)

        if value_before is None:
            raise KeyError(f"Key path '{self.file_change.key_path}' does not exist in {self.path}")
        elif not contains_pattern(search_for, value_before) and not self.file_change.ignore_missing_version:
            raise ValueError(
                f"Key '{self.file_change.key_path}' in {self.path} does not contain the correct contents: "
                f"{raw_search_pattern}"
            )

        new_value = search_for.sub(replace_with, value_before)
        log_changes(f"{self.path}:{self.file_change.key_path}", value_before, new_value, dry_run)

        if dry_run:
            return

        set_nested_value(toml_data, new_value, self.file_change.key_path)

        self.path.write_text(tomlkit.dumps(toml_data), encoding="utf-8")



"""Helper functions for the config module."""

from __future__ import annotations

import re
from typing import Dict, List

from wcmatch import glob

from bumpversion.config.models import FileChange
from bumpversion.exceptions import BumpVersionError
from bumpversion.versioning.models import VersionComponentSpec


def get_all_file_configs(config_dict: dict) -> List[FileChange]:
    """Make sure all version parts are included."""
    defaults = {
        "parse": config_dict["parse"],
        "serialize": config_dict["serialize"],
        "search": config_dict["search"],
        "replace": config_dict["replace"],
        "ignore_missing_version": config_dict["ignore_missing_version"],
        "ignore_missing_file": config_dict["ignore_missing_files"],
        "regex": config_dict["regex"],
    }
    files = [{k: v for k, v in filecfg.items() if v is not None} for filecfg in config_dict["files"]]
    for f in files:
        f.update({k: v for k, v in defaults.items() if k not in f})
    return [FileChange(**f) for f in files]


def get_all_part_configs(config_dict: dict) -> Dict[str, VersionComponentSpec]:
    """Make sure all version parts are included."""
    try:
        parsing_groups = list(re.compile(config_dict["parse"]).groupindex.keys())
    except re.error as e:
        raise BumpVersionError(f"Could not parse regex '{config_dict['parse']}': {e}") from e
    parts = config_dict["parts"]

    part_configs = {}
    for label in parsing_groups:
        is_independent = label.startswith("$")
        part_configs[label] = (
            VersionComponentSpec(**parts[label])
            if label in parts
            else VersionComponentSpec(independent=is_independent)
        )
    return part_configs


def resolve_glob_files(file_cfg: FileChange) -> List[FileChange]:
    """
    Return a list of file configurations that match the glob pattern.

    Args:
        file_cfg: The file configuration containing the glob pattern

    Returns:
        A list of resolved file configurations according to the pattern.
    """
    files: List[FileChange] = []
    exclude = file_cfg.glob_exclude or []
    glob_flags = glob.GLOBSTAR | glob.FORCEUNIX | glob.SPLIT
    for filename_glob in glob.glob(file_cfg.glob, flags=glob_flags, exclude=exclude):
        new_file_cfg = file_cfg.model_copy()
        new_file_cfg.filename = filename_glob
        new_file_cfg.glob = None
        files.append(new_file_cfg)
    return files



"""
Automatically detect the true Python type of a string and cast it to the correct type.

Based on https://github.com/cgreer/cgAutoCast/blob/master/cgAutoCast.py

Only used by Legacy configuration file parser.
"""

import contextlib
from typing import Any


def boolify(s: str) -> bool:
    """Convert a string to a boolean."""
    if s in {"True", "true"}:
        return True
    if s in {"False", "false"}:
        return False
    raise ValueError("Not Boolean Value!")


def noneify(s: str) -> None:
    """Convert a string to None."""
    if s == "None":
        return None
    raise ValueError("Not None Value!")


def listify(s: str) -> list:
    """
    Convert a string representation of a list into list of homogeneous basic types.

    Type of elements in list is determined via first element. Successive elements are
    cast to that type.

    Args:
        s: String representation of a list.

    Raises:
        ValueError: If string does not represent a list.
        TypeError: If string does not represent a list of homogeneous basic types.

    Returns:
        List of homogeneous basic types.
    """
    if "\n" in s:
        str_list = s.strip().split("\n")
    elif "," in s:
        str_list = s.strip().split(",")
    else:
        raise ValueError("Not a List")

    # derive the type of the variable
    element_caster = str
    for caster in (boolify, int, float, noneify, element_caster):
        with contextlib.suppress(ValueError):
            caster(str_list[0])  # type: ignore[operator]
            element_caster = caster  # type: ignore[assignment]
            break
    # cast all elements
    try:
        return [element_caster(x) for x in str_list]
    except ValueError as e:
        raise TypeError("Autocasted list must be all same type") from e


def autocast_value(var: Any) -> Any:
    """
    Guess the string representation of the variable's type.

    Args:
        var: Value to autocast.

    Returns:
        The autocasted value.
    """
    if not isinstance(var, str):  # don't need to guess non-string types
        return var

    # guess string representation of var
    for caster in (boolify, int, float, noneify, listify):
        with contextlib.suppress(ValueError):
            return caster(var)  # type: ignore[operator]

    return var



"""Tests for `bumpversion` package."""

import shutil
import traceback
from pathlib import Path

from click.testing import Result

from bumpversion import cli
from tests.conftest import inside_dir

TEST_REPLACE_CONFIG = {
    "tool": {
        "bumpversion": {
            "allow_dirty": True,
            "commit": False,
            "current_version": "2.17.7",
            "files": [],
            "message": "Bump version: {current_version} → {new_version}",
            "parse": "(?P<major>\\d+)\\.(?P<minor>\\d+)\\.(?P<patch>\\d+)",
            "parts": {
                "major": {},
                "minor": {},
                "patch": {},
            },
            "replace": "{new_version}",
            "search": "{current_version}",
            "serialize": ["{major}.{minor}.{patch}"],
            "sign_tags": False,
            "tag": False,
            "tag_message": "Bump version: {current_version} → {new_version}",
            "tag_name": "v{new_version}",
        }
    }
}


class TestReplaceCLI:
    """Test the replace CLI subcommand."""

    class TestDefaultReplacesVersion:
        """Test the default behavior of the replace subcommand."""

        def test_default_affects_all_configured_files(self, mocker, tmp_path, fixtures_path, runner):
            """The replace subcommand should replace the version in all configured files."""
            # Arrange
            toml_path = fixtures_path / "basic_cfg.toml"
            config_path = tmp_path / "pyproject.toml"
            shutil.copy(toml_path, config_path)

            mocked_modify_files = mocker.patch("bumpversion.cli.modify_files")
            with inside_dir(tmp_path):
                result: Result = runner.invoke(cli.cli, ["replace", "--new-version", "1.1.0"])

            if result.exit_code != 0:
                print("Here is the output:")
                print(result.output)
                print(traceback.print_exception(result.exc_info[1]))

            assert result.exit_code == 0

            call_args = mocked_modify_files.call_args[0]
            configured_files = call_args[0]
            assert len(configured_files) == 3
            actual_filenames = {f.file_change.filename for f in configured_files}
            assert actual_filenames == {"setup.py", "CHANGELOG.md", "bumpversion/__init__.py"}

        def test_can_limit_to_specific_files(self, mocker, git_repo, fixtures_path, runner):
            """The replace subcommand should set the files to only the specified files."""
            # Arrange
            toml_path = fixtures_path / "basic_cfg.toml"
            config_path = git_repo / "pyproject.toml"
            shutil.copy(toml_path, config_path)

            mocked_modify_files = mocker.patch("bumpversion.cli.modify_files")
            with inside_dir(git_repo):
                result: Result = runner.invoke(cli.cli, ["replace", "--no-configured-files", "VERSION"])

            if result.exit_code != 0:
                print("Here is the output:")
                print(result.output)
                print(traceback.print_exception(result.exc_info[1]))

            assert result.exit_code == 0

            call_args = mocked_modify_files.call_args[0]
            configured_files = call_args[0]
            assert len(configured_files) == 1
            assert configured_files[0].file_change.filename == "VERSION"

    class TestOptions:
        """Test the options of the replace subcommand."""

        def test_missing_newversion_is_set_to_none(self, mocker, tmp_path, fixtures_path, runner):
            """The replace subcommand should set new_version to None in the context."""
            # Arrange
            toml_path = fixtures_path / "basic_cfg.toml"
            config_path = tmp_path / "pyproject.toml"
            shutil.copy(toml_path, config_path)

            mocked_modify_files = mocker.patch("bumpversion.cli.modify_files")
            with inside_dir(tmp_path):
                result: Result = runner.invoke(cli.cli, ["replace"])

            if result.exit_code != 0:
                print("Here is the output:")
                print(result.output)
                print(traceback.print_exception(result.exc_info[1]))

            assert result.exit_code == 0

            call_args = mocked_modify_files.call_args[0]
            assert call_args[2] is None

        def test_ignores_missing_files_with_option(self, mocker, tmp_path, fixtures_path, runner):
            """The replace subcommand should ignore missing."""

            config_file = tmp_path / ".bumpversion.toml"
            config_file.write_text(
                "[tool.bumpversion]\n"
                'current_version = "0.0.1"\n'
                "allow_dirty = true\n\n"
                "[[tool.bumpversion.files]]\n"
                'filename = "VERSION"\n'
                "regex = false\n",
                encoding="utf-8",
            )

            # Act
            with inside_dir(tmp_path):
                result: Result = runner.invoke(
                    cli.cli,
                    [
                        "replace",
                        "--verbose",
                        "--no-regex",
                        "--no-configured-files",
                        "--ignore-missing-files",
                        "VERSION",
                    ],
                )

            # Assert
            if result.exit_code != 0:
                print("Here is the output:")
                print(result.output)
                print(traceback.print_exception(result.exc_info[1]))

            assert result.exit_code == 0

    class TestSearchInputs:
        """Test the search inputs of the replace subcommand."""

        def test_accepts_plain_string(self, tmp_path, fixtures_path, runner):
            """Replace should not worry if the search values have version info."""
            from tomlkit import dumps

            # Arrange
            config_path = tmp_path / "pyproject.toml"
            config_path.write_text(dumps(TEST_REPLACE_CONFIG), encoding="utf-8")
            doc_path = tmp_path / "docs.yaml"
            doc_path.write_text(
                "url: https://github.com/sampleuser/workflows/main/.github/update_mailmap.py", encoding="utf-8"
            )

            with inside_dir(tmp_path):
                result: Result = runner.invoke(
                    cli.cli,
                    [
                        "replace",
                        "--no-configured-files",
                        "--search",
                        "/workflows/main/",
                        "--replace",
                        "/workflows/v{current_version}/",
                        "./docs.yaml",
                    ],
                )

            if result.exit_code != 0:
                print("Here is the output:")
                print(result.output)
                print(traceback.print_exception(result.exc_info[1]))

            assert result.exit_code == 0

        def test_unintentional_valid_regex_still_found(self, tmp_path: Path, caplog, runner) -> None:
            """A search string not meant to be a regex (but is) is still found and replaced correctly."""
            # Arrange
            search = "(unreleased)"
            replace = "(2023-01-01)"

            version_path = tmp_path / "VERSION"
            version_path.write_text(
                "# Changelog\n\n## [0.0.1 (unreleased)](https://cool.url)\n\n- Test unreleased package.\n",
                encoding="utf-8",
            )
            config_file = tmp_path / ".bumpversion.toml"
            config_file.write_text(
                "[tool.bumpversion]\n"
                'current_version = "0.0.1"\n'
                "allow_dirty = true\n\n"
                "[[tool.bumpversion.files]]\n"
                'filename = "VERSION"\n'
                "regex = false\n"
                f'search = "{search}"\n'
                f'replace = "{replace}"\n',
                encoding="utf-8",
            )

            # Act
            with inside_dir(tmp_path):
                result: Result = runner.invoke(
                    cli.cli,
                    [
                        "replace",
                        "--verbose",
                        "--no-regex",
                        "--no-configured-files",
                        "--search",
                        search,
                        "--replace",
                        replace,
                        "VERSION",
                    ],
                )

            # Assert
            if result.exit_code != 0:
                print("Here is the output:")
                print(result.output)
                print(traceback.print_exception(result.exc_info[1]))

            assert result.exit_code == 0
            assert (
                version_path.read_text()
                == "# Changelog\n\n## [0.0.1 (2023-01-01)](https://cool.url)\n\n- Test unreleased package.\n"
            )

    class TestReplaceInputs:
        """Test the replace inputs of the replace subcommand."""

        def test_accepts_empty_string(self, tmp_path, fixtures_path, runner):
            """Replace should be able to replace strings with an empty string."""
            from tomlkit import dumps

            # Arrange
            config_path = tmp_path / "pyproject.toml"
            config_path.write_text(dumps(TEST_REPLACE_CONFIG), encoding="utf-8")
            doc_path = tmp_path / "docs.yaml"
            doc_path.write_text("We should censor profanity\n\n", encoding="utf-8")

            with inside_dir(tmp_path):
                result: Result = runner.invoke(
                    cli.cli,
                    [
                        "replace",
                        "--verbose",
                        "--no-configured-files",
                        "--allow-dirty",
                        "--search",
                        "profanity",
                        "--replace",
                        "",
                        "./docs.yaml",
                    ],
                )

            if result.exit_code != 0:
                print("Here is the output:")
                print(result.output)
                print(traceback.print_exception(result.exc_info[1]))
            assert result.exit_code == 0
            assert doc_path.read_text() == "We should censor \n\n"

        def test_accepts_plain_string(self, tmp_path, fixtures_path, runner):
            """Replace should not worry if the replace values have version info."""
            from tomlkit import dumps

            # Arrange
            config_path = tmp_path / "pyproject.toml"
            config_path.write_text(dumps(TEST_REPLACE_CONFIG), encoding="utf-8")
            doc_path = tmp_path / "docs.yaml"
            doc_path.write_text(
                "url: https://github.com/sampleuser/workflows/v2.17.7/.github/update_mailmap.py", encoding="utf-8"
            )

            with inside_dir(tmp_path):
                result: Result = runner.invoke(
                    cli.cli,
                    [
                        "replace",
                        "--no-configured-files",
                        "--search",
                        "/workflows/v{current_version}/",
                        "--replace",
                        "/workflows/main/",
                        "./docs.yaml",
                    ],
                )

            if result.exit_code != 0:
                print("Here is the output:")
                print(result.output)
                print(traceback.print_exception(result.exc_info[1]))

            assert result.exit_code == 0
            assert (
                doc_path.read_text() == "url: https://github.com/sampleuser/workflows/main/.github/update_mailmap.py"
            )



"""Context for rendering messages and tags."""

import calendar
import datetime
from collections import ChainMap
from dataclasses import asdict
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:  # pragma: no-coverage
    from bumpversion.config import Config
    from bumpversion.scm import SCMInfo
    from bumpversion.versioning.models import Version

MONTH_ABBREVIATIONS = [abbr for abbr in calendar.month_abbr if abbr]

CALVER_PATTERN_MAP = {
    "{YYYY}": r"(?:[1-9][0-9]{3})",
    "{YY}": r"(?:[1-9][0-9]?)",
    "{0Y}": r"(?:[0-9]{2})",
    "{MMM}": f'(?:{"|".join(MONTH_ABBREVIATIONS)})',
    "{MM}": r"(?:1[0-2]|[1-9])",
    "{0M}": r"(?:1[0-2]|0[1-9])",
    "{DD}": r"(?:3[0-1]|[1-2][0-9]|[1-9])",
    "{0D}": r"(?:3[0-1]|[1-2][0-9]|0[1-9])",
    "{JJJ}": r"(?:36[0-6]|3[0-5][0-9]|[1-2][0-9][0-9]|[1-9][0-9]|[1-9])",
    "{00J}": r"(?:36[0-6]|3[0-5][0-9]|[1-2][0-9][0-9]|0[1-9][0-9]|00[1-9])",
    "{Q}": r"(?:[1-4])",
    "{WW}": r"(?:5[0-3]|[1-4][0-9]|[0-9])",
    "{0W}": r"(?:5[0-3]|[0-4][0-9])",
    "{UU}": r"(?:5[0-3]|[1-4][0-9]|[0-9])",
    "{0U}": r"(?:5[0-3]|[0-4][0-9])",
    "{VV}": r"(?:5[0-3]|[1-4][0-9]|[1-9])",
    "{0V}": r"(?:5[0-3]|[1-4][0-9]|0[1-9])",
    "{GGGG}": r"(?:[1-9][0-9]{3})",
    "{GG}": r"(?:[0-9][0-9]?)",
    "{0G}": r"(?:[0-9]{2})",
    "{INC0}": r"(?:[0-9]+)",
    "{INC1}": r"[1-9][0-9]*",
}


def calver_string_to_regex(calver_format: str) -> str:
    """Convert the calver format string to a regex pattern."""
    pattern = calver_format
    for key, value in CALVER_PATTERN_MAP.items():
        pattern = pattern.replace(key, value)
    return pattern


def prefixed_environ() -> dict:
    """Return a dict of the environment with keys wrapped in `${}`."""
    import os

    return {f"${key}": value for key, value in os.environ.items()}


def base_context(scm_info: Optional["SCMInfo"] = None) -> ChainMap:
    """The default context for rendering messages and tags."""
    from bumpversion.scm import SCMInfo  # Including this here to avoid circular imports

    scm = asdict(scm_info) if scm_info else asdict(SCMInfo())

    return ChainMap(
        {
            "now": datetime.datetime.now(),
            "utcnow": datetime.datetime.now(datetime.timezone.utc),
        },
        prefixed_environ(),
        scm,
        {c: c for c in ("#", ";")},
    )


def get_context(
    config: "Config", current_version: Optional["Version"] = None, new_version: Optional["Version"] = None
) -> ChainMap:
    """Return the context for rendering messages and tags."""
    ctx = base_context(config.scm_info)
    ctx = ctx.new_child({"current_version": config.current_version})
    if current_version:
        ctx = ctx.new_child({f"current_{part}": current_version[part].value for part in current_version})
    if new_version:
        ctx = ctx.new_child({f"new_{part}": new_version[part].value for part in new_version})
    return ctx


def get_datetime_info(current_dt: datetime.datetime) -> dict:
    """Return the full structure of the given datetime for formatting."""
    return {
        "YYYY": current_dt.strftime("%Y"),
        "YY": current_dt.strftime("%y").lstrip("0") or "0",
        "0Y": current_dt.strftime("%y"),
        "MMM": current_dt.strftime("%b"),
        "MM": str(current_dt.month),
        "0M": current_dt.strftime("%m"),
        "DD": str(current_dt.day),
        "0D": current_dt.strftime("%d"),
        "JJJ": current_dt.strftime("%j").lstrip("0"),
        "00J": current_dt.strftime("%j"),
        "Q": str((current_dt.month - 1) // 3 + 1),
        "WW": current_dt.strftime("%W").lstrip("0") or "0",
        "0W": current_dt.strftime("%W"),
        "UU": current_dt.strftime("%U").lstrip("0") or "0",
        "0U": current_dt.strftime("%U"),
        "VV": current_dt.strftime("%V").lstrip("0") or "0",
        "0V": current_dt.strftime("%V"),
        "GGGG": current_dt.strftime("%G"),
        "GG": current_dt.strftime("%G")[2:].lstrip("0") or "0",
        "0G": current_dt.strftime("%G")[2:],
    }



"""Tests of the VersionSpec model."""

import pytest
from bumpversion.versioning.models import VersionSpec
from bumpversion.versioning.models import VersionComponentSpec


class TestVersionSpec:
    """Tests for the version spec."""

    class TestCreation:
        """Tests for creating a version spec."""

        def test_empty_components_raises_error(self):
            """An empty components dict raises an error."""
            with pytest.raises(ValueError):
                VersionSpec({}, [])

        class TestComponentOrder:
            def test_empty_order_uses_order_of_components(self):
                """If the order is empty, it uses the component order."""
                # Arrange
                config = {
                    "major": VersionComponentSpec(),
                    "minor": VersionComponentSpec(),
                    "patch": VersionComponentSpec(),
                }

                # Act
                version_spec = VersionSpec(config, [])

                # Assert
                assert version_spec.order == ["major", "minor", "patch"]

            def test_extra_items_raises_error(self):
                """If the order contains component names that do not exist, it raises and error."""
                # Arrange
                config = {
                    "major": VersionComponentSpec(),
                    "minor": VersionComponentSpec(),
                    "patch": VersionComponentSpec(),
                }

                # Act
                with pytest.raises(ValueError):
                    VersionSpec(config, ["major", "minor", "patch", "build"])

            def test_subset_of_items_works_fine(self):
                """An order containing a subset of component names works fine."""
                # Arrange
                config = {
                    "major": VersionComponentSpec(),
                    "minor": VersionComponentSpec(),
                    "patch": VersionComponentSpec(),
                }

                # Act
                version_spec = VersionSpec(config, ["major", "patch"])

                # Assert
                assert dict(version_spec.dependency_map) == {"major": ["patch"]}

        def test_dependency_map_follows_order(self):
            """The order of components correctly creates the dependency map."""
            # Arrange
            config = {
                "major": VersionComponentSpec(),
                "minor": VersionComponentSpec(),
                "patch": VersionComponentSpec(),
            }

            # Act
            version_spec = VersionSpec(config, ["major", "minor", "patch"])
            version_spec2 = VersionSpec(config, ["patch", "minor", "major"])

            # Assert
            assert dict(version_spec.dependency_map) == {"major": ["minor"], "minor": ["patch"]}
            assert dict(version_spec2.dependency_map) == {"patch": ["minor"], "minor": ["major"]}

        def test_dependency_map_skips_independent_components(self):
            """Independent components are not in the dependency map."""
            # Arrange
            config = {
                "major": VersionComponentSpec(),
                "minor": VersionComponentSpec(independent=True),
                "patch": VersionComponentSpec(),
            }

            # Act
            version_spec = VersionSpec(config, ["major", "minor", "patch"])

            # Assert
            assert dict(version_spec.dependency_map) == {"major": ["patch"]}

    class TestGenerateVersion:
        """Tests of the generate_version method."""

        def test_empty_values_creates_default_version(self):
            """An empty values dict raises an error."""
            # Arrange
            config = {
                "major": VersionComponentSpec(),
                "minor": VersionComponentSpec(),
                "patch": VersionComponentSpec(),
                "build": VersionComponentSpec(independent=True),
            }
            version_spec = VersionSpec(config)

            # Act
            version = version_spec.create_version({})

            # Assert
            assert version["major"].value == "0"
            assert version["minor"].value == "0"
            assert version["patch"].value == "0"
            assert version["build"].value == "0"

        def test_missing_values_uses_component_first_value(self):
            """A missing value raises an error."""
            # Arrange
            config = {
                "major": VersionComponentSpec(),
                "minor": VersionComponentSpec(),
                "patch": VersionComponentSpec(),
                "build": VersionComponentSpec(independent=True),
            }
            version_spec = VersionSpec(config)

            # Act
            version = version_spec.create_version({"major": "1", "minor": "2"})

            # Assert
            assert version["major"].value == "1"
            assert version["minor"].value == "2"
            assert version["patch"].value == "0"
            assert version["build"].value == "0"

        def test_extra_values_ignored(self):
            """Extra values are ignored."""
            # Arrange
            config = {
                "major": VersionComponentSpec(),
                "minor": VersionComponentSpec(),
                "patch": VersionComponentSpec(),
            }
            version_spec = VersionSpec(config)

            # Act
            version = version_spec.create_version({"major": "1", "minor": "2", "patch": "3", "build": "4"})

            # Assert
            assert version["major"].value == "1"
            assert version["minor"].value == "2"
            assert version["patch"].value == "3"
            assert "build" not in version

    class TestGetDependents:
        """Tests of the get_dependents method."""

        def test_bad_component_name_returns_empty_list(self):
            """Getting the dependents of a non-existing component returns an empty list."""
            # Arrange
            config = {
                "major": VersionComponentSpec(),
                "minor": VersionComponentSpec(),
                "patch": VersionComponentSpec(),
            }
            version_spec = VersionSpec(config)

            # Act
            dependents = version_spec.get_dependents("invalid")

            # Assert
            assert dependents == []

        def test_extra_values_ignored(self):
            """Extra values are ignored."""
            # Arrange
            config = {
                "major": VersionComponentSpec(),
                "minor": VersionComponentSpec(),
                "patch": VersionComponentSpec(),
            }
            version_spec = VersionSpec(config)

            # Act
            patch_dependents = version_spec.get_dependents("patch")
            minor_dependents = version_spec.get_dependents("minor")
            major_dependents = version_spec.get_dependents("major")

            # Assert
            assert patch_dependents == []
            assert minor_dependents == ["patch"]
            assert major_dependents == ["minor", "patch"]
